"""scripts/click_types.py

Custom click parameter types and utilities.
"""

from typing import List, Any, Tuple, Dict
import pathlib
import glob
import re
import os
import string

import click


class GlobbityGlob(click.ParamType):
    """Expands a glob pattern to Path objects"""

    name = "glob"

    def convert(self, value: str, *args: Any) -> List[pathlib.Path]:
        return [pathlib.Path(f) for f in glob.glob(value)]


class PathlibPath(click.Path):
    """Converts a string to a pathlib.Path object"""

    def convert(self, *args: Any) -> pathlib.Path:
        return pathlib.Path(str(super().convert(*args)))


RasterPatternType = Tuple[List[str], Dict[Tuple[str, ...], str]]


def _parse_raster_pattern(raster_pattern: str) -> Tuple[List[str], str, str]:
    """Parse a raster pattern string using Python format syntax.

    Extracts names of unique placeholders, a glob pattern
    and a regular expression to retrieve files matching the given pattern.

    Example:

        >>> _parse_raster_pattern('{key1}/{key2}_{}.tif')
        (['key1', 'key2'], '*/*_*.tif', '(?P<key1>[^\\W_]+)/(?P<key2>[^\\W_]+)_.*?\\.tif')

    """

    # raises ValueError on invalid patterns
    parsed_value = string.Formatter().parse(raster_pattern)

    keys: List[str] = []
    glob_pattern: List[str] = []
    regex_pattern: List[str] = []

    for before_field, field_name, _, _ in parsed_value:
        glob_pattern += before_field
        regex_pattern += re.escape(before_field)

        if field_name is None:
            # no placeholder
            continue

        glob_pattern.append("*")

        if field_name == "":
            # unnamed placeholder
            regex_pattern.append(".*?")
        elif field_name in keys:
            # duplicate placeholder
            key_group_number = keys.index(field_name) + 1
            regex_pattern.append(rf"\{key_group_number}")
        else:
            # new placeholder
            keys.append(field_name)
            regex_pattern += rf"(?P<{field_name}>[^\W_]+)"

    return keys, "".join(glob_pattern), "".join(regex_pattern)


class RasterPattern(click.ParamType):
    """Expands a pattern following the Python format specification to matching files"""

    name = "raster-pattern"

    def convert(self, value: str, *args: Any) -> RasterPatternType:
        value = os.path.abspath(value)
        try:
            keys, glob_pattern, regex_pattern = _parse_raster_pattern(value)
        except ValueError as exc:
            self.fail(f"Invalid pattern: {exc!s}")

        if not keys:
            self.fail("Pattern must contain at least one placeholder")

        if not all(re.match(r"\w", key) for key in keys):
            self.fail("Key names must be alphanumeric")

        # use glob to find candidates, regex to extract placeholder values
        candidates = (os.path.abspath(c) for c in glob.glob(glob_pattern))
        matched_candidates = [
            re.match(regex_pattern, candidate) for candidate in candidates
        ]

        if not any(matched_candidates):
            self.fail("Given pattern matches no files")

        key_combinations = [
            tuple(match.groups()) for match in matched_candidates if match
        ]
        if len(key_combinations) != len(set(key_combinations)):
            self.fail("Pattern leads to duplicate keys")

        files = {
            tuple(match.groups()): match.group(0)
            for match in matched_candidates
            if match
        }
        return keys, files


class TOMLFile(click.ParamType):
    """Parses a TOML file to a dict"""

    name = "toml-file"

    def convert(self, value: str, *args: Any) -> Dict[str, Any]:
        import toml

        return dict(toml.load(value))


class Hostname(click.ParamType):
    """Parses a string to a valid hostname"""

    name = "url"

    def __init__(self, default_port: int = 5000, default_scheme: str = "http") -> None:
        self.default_port = default_port
        self.default_scheme = default_scheme

    def convert(self, value: str, *args: Any) -> str:
        from urllib.parse import urlparse, urlunparse

        parsed_url = urlparse(value)

        if not parsed_url.netloc:
            value_with_scheme = "://".join([self.default_scheme, value])
            parsed_url = urlparse(value_with_scheme)

        # remove everything we don't need
        return urlunparse(
            [parsed_url.scheme, parsed_url.netloc, parsed_url.path, "", "", ""]
        )



import pytest

import numpy as np
import rasterio
import rasterio.features
from shapely.geometry import shape, MultiPolygon


def geometry_mismatch(shape1, shape2):
    """Compute relative mismatch of two shapes"""
    return shape1.symmetric_difference(shape2).area / shape1.union(shape2).area


def convex_hull_exact(src):
    kwargs = dict(bidx=1, band=False, as_mask=True, geographic=True)

    data = src.read()
    if np.any(np.isnan(data)) and src.nodata is not None:
        # hack: replace NaNs with nodata to make sure they are excluded
        with rasterio.MemoryFile() as memfile, memfile.open(**src.profile) as tmpsrc:
            data[np.isnan(data)] = src.nodata
            tmpsrc.write(data)
            dataset_shape = list(rasterio.features.dataset_features(tmpsrc, **kwargs))
    else:
        dataset_shape = list(rasterio.features.dataset_features(src, **kwargs))

    convex_hull = MultiPolygon(
        [shape(s["geometry"]) for s in dataset_shape]
    ).convex_hull
    return convex_hull


@pytest.mark.parametrize("large_raster_threshold", [None, 0])
@pytest.mark.parametrize("use_chunks", [True, False, None])
@pytest.mark.parametrize("nodata_type", ["nodata", "masked", "none", "nan"])
def test_compute_metadata(
    big_raster_file_nodata,
    big_raster_file_nomask,
    big_raster_file_mask,
    raster_file_float,
    nodata_type,
    use_chunks,
    large_raster_threshold,
):
    from terracotta import raster

    if nodata_type == "nodata":
        raster_file = big_raster_file_nodata
    elif nodata_type == "masked":
        raster_file = big_raster_file_mask
    elif nodata_type == "none":
        raster_file = big_raster_file_nomask
    elif nodata_type == "nan":
        raster_file = raster_file_float

    if use_chunks:
        pytest.importorskip("crick")

    with rasterio.open(str(raster_file)) as src:
        data = src.read(1, masked=True)
        valid_data = np.ma.masked_invalid(data).compressed()
        convex_hull = convex_hull_exact(src)

    # compare
    if nodata_type == "none":
        with pytest.warns(UserWarning) as record:
            mtd = raster.compute_metadata(
                str(raster_file),
                use_chunks=use_chunks,
                large_raster_threshold=large_raster_threshold,
            )
            assert "does not have a valid nodata value" in str(record[0].message)
    else:
        mtd = raster.compute_metadata(
            str(raster_file),
            use_chunks=use_chunks,
            large_raster_threshold=large_raster_threshold,
        )

    np.testing.assert_allclose(
        mtd["valid_percentage"], 100 * valid_data.size / data.size
    )
    np.testing.assert_allclose(mtd["range"], (valid_data.min(), valid_data.max()))
    np.testing.assert_allclose(mtd["mean"], valid_data.mean())
    np.testing.assert_allclose(mtd["stdev"], valid_data.std())

    # allow some error margin since we only compute approximate quantiles
    np.testing.assert_allclose(
        mtd["percentiles"],
        np.percentile(valid_data, np.arange(1, 100)),
        rtol=2e-2,
        atol=valid_data.max() / 100,
    )

    assert geometry_mismatch(shape(mtd["convex_hull"]), convex_hull) < 1e-6


@pytest.mark.parametrize("nodata_type", ["nodata", "masked", "none", "nan"])
def test_compute_metadata_approximate(
    nodata_type,
    big_raster_file_nodata,
    big_raster_file_mask,
    big_raster_file_nomask,
    raster_file_float,
):
    from terracotta import raster

    if nodata_type == "nodata":
        raster_file = big_raster_file_nodata
    elif nodata_type == "masked":
        raster_file = big_raster_file_mask
    elif nodata_type == "none":
        raster_file = big_raster_file_nomask
    elif nodata_type == "nan":
        raster_file = raster_file_float

    with rasterio.open(str(raster_file)) as src:
        data = src.read(1, masked=True)
        valid_data = np.ma.masked_invalid(data).compressed()
        convex_hull = convex_hull_exact(src)

    # compare
    if nodata_type == "none":
        with pytest.warns(UserWarning) as record:
            mtd = raster.compute_metadata(str(raster_file), max_shape=(512, 512))
            assert "does not have a valid nodata value" in str(record[0].message)
    else:
        mtd = raster.compute_metadata(str(raster_file), max_shape=(512, 512))

    np.testing.assert_allclose(
        mtd["valid_percentage"], 100 * valid_data.size / data.size, atol=1
    )
    np.testing.assert_allclose(
        mtd["range"], (valid_data.min(), valid_data.max()), atol=valid_data.max() / 100
    )
    np.testing.assert_allclose(mtd["mean"], valid_data.mean(), rtol=0.02)
    np.testing.assert_allclose(mtd["stdev"], valid_data.std(), rtol=0.02)

    np.testing.assert_allclose(
        mtd["percentiles"],
        np.percentile(valid_data, np.arange(1, 100)),
        atol=valid_data.max() / 100,
        rtol=0.02,
    )

    assert geometry_mismatch(shape(mtd["convex_hull"]), convex_hull) < 0.05


def test_compute_metadata_invalid_options(big_raster_file_nodata):
    from terracotta import raster

    with pytest.raises(ValueError):
        raster.compute_metadata(
            str(big_raster_file_nodata), max_shape=(256, 256), use_chunks=True
        )

    with pytest.raises(ValueError):
        raster.compute_metadata(str(big_raster_file_nodata), max_shape=(256, 256, 1))


@pytest.mark.parametrize("use_chunks", [True, False])
def test_compute_metadata_invalid_raster(invalid_raster_file, use_chunks):
    from terracotta import raster

    if use_chunks:
        pytest.importorskip("crick")

    with pytest.raises(ValueError):
        raster.compute_metadata(str(invalid_raster_file), use_chunks=use_chunks)


def test_compute_metadata_nocrick(big_raster_file_nodata, monkeypatch):
    with rasterio.open(str(big_raster_file_nodata)) as src:
        data = src.read(1, masked=True)
        valid_data = np.ma.masked_invalid(data).compressed()
        convex_hull = convex_hull_exact(src)

    from terracotta import exceptions
    import terracotta.drivers.geotiff_raster_store

    with monkeypatch.context() as m:
        m.setattr(terracotta.raster, "has_crick", False)

        with pytest.warns(exceptions.PerformanceWarning):
            mtd = terracotta.drivers.geotiff_raster_store.raster.compute_metadata(
                str(big_raster_file_nodata), use_chunks=True
            )

    # compare
    np.testing.assert_allclose(
        mtd["valid_percentage"], 100 * valid_data.size / data.size
    )
    np.testing.assert_allclose(mtd["range"], (valid_data.min(), valid_data.max()))
    np.testing.assert_allclose(mtd["mean"], valid_data.mean())
    np.testing.assert_allclose(mtd["stdev"], valid_data.std())

    # allow error of 1%, since we only compute approximate quantiles
    np.testing.assert_allclose(
        mtd["percentiles"], np.percentile(valid_data, np.arange(1, 100)), rtol=2e-2
    )

    assert geometry_mismatch(shape(mtd["convex_hull"]), convex_hull) < 1e-6


def test_compute_metadata_unoptimized(unoptimized_raster_file):
    from terracotta import exceptions
    from terracotta import raster

    with rasterio.open(str(unoptimized_raster_file)) as src:
        data = src.read(1, masked=True)
        valid_data = np.ma.masked_invalid(data).compressed()
        convex_hull = convex_hull_exact(src)

    # compare
    with pytest.warns(exceptions.PerformanceWarning):
        mtd = raster.compute_metadata(str(unoptimized_raster_file), use_chunks=False)

    np.testing.assert_allclose(
        mtd["valid_percentage"], 100 * valid_data.size / data.size
    )
    np.testing.assert_allclose(mtd["range"], (valid_data.min(), valid_data.max()))
    np.testing.assert_allclose(mtd["mean"], valid_data.mean())
    np.testing.assert_allclose(mtd["stdev"], valid_data.std())

    # allow some error margin since we only compute approximate quantiles
    np.testing.assert_allclose(
        mtd["percentiles"], np.percentile(valid_data, np.arange(1, 100)), rtol=2e-2
    )

    assert geometry_mismatch(shape(mtd["convex_hull"]), convex_hull) < 1e-6


@pytest.mark.parametrize("preserve_values", [True, False])
@pytest.mark.parametrize("resampling_method", ["nearest", "linear", "cubic", "average"])
def test_get_raster_tile(raster_file, preserve_values, resampling_method):
    from terracotta import raster

    data = raster.get_raster_tile(
        str(raster_file),
        reprojection_method=resampling_method,
        resampling_method=resampling_method,
        preserve_values=preserve_values,
        tile_size=(256, 256),
    )
    assert data.shape == (256, 256)


def test_get_raster_tile_out_of_bounds(raster_file):
    from terracotta import exceptions
    from terracotta import raster

    bounds = (
        -1e30,
        -1e30,
        1e30,
        1e30,
    )

    with pytest.raises(exceptions.TileOutOfBoundsError):
        raster.get_raster_tile(str(raster_file), tile_bounds=bounds)


def test_get_raster_no_nodata(big_raster_file_nomask):
    from terracotta import raster

    tile_size = (256, 256)
    out = raster.get_raster_tile(str(big_raster_file_nomask), tile_size=tile_size)
    assert out.shape == tile_size


def test_invalid_resampling_method():
    from terracotta import raster

    with pytest.raises(ValueError) as exc:
        raster.get_resampling_enum("not-a-resampling-method")
    assert "unknown resampling method" in str(exc)



from typing import Any, cast, Callable, Type, TYPE_CHECKING
import copy

from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin
from apispec_webframeworks.flask import FlaskPlugin

from flask import Flask, Blueprint, current_app, send_file, jsonify
from flask_cors import CORS

import marshmallow

from terracotta import exceptions, __version__

# define blueprints, will be populated by submodules
TILE_API = Blueprint("tile_api", "terracotta.server")
METADATA_API = Blueprint("metadata_api", "terracotta.server")
SPEC_API = Blueprint("spec_api", "terracotta.server")

# create an APISpec
SPEC = APISpec(
    title="Terracotta",
    version=__version__,
    openapi_version="3.0",
    info=dict(description="A modern XYZ Tile Server in Python"),
    plugins=[FlaskPlugin(), MarshmallowPlugin()],
    # ensure that endpoints are interpreted as relative paths
    # see Redocly/redoc#1172
    servers=[dict(url="")],
)


def _abort(status_code: int, message: str = "") -> Any:
    response = jsonify({"message": message})
    response.status_code = status_code
    return response


def _setup_error_handlers(app: Flask) -> None:
    def register_error_handler(
        exc: Type[Exception], func: Callable[[Exception], Any]
    ) -> None:
        if TYPE_CHECKING:  # pragma: no cover
            # Flask defines this type only during type checking
            from flask.typing import ErrorHandlerCallable

            func = cast(ErrorHandlerCallable, func)

        app.register_error_handler(exc, func)

    def handle_tile_out_of_bounds_error(exc: Exception) -> Any:
        # send empty image
        from terracotta import get_settings, image

        settings = get_settings()
        return send_file(
            image.empty_image(settings.DEFAULT_TILE_SIZE), mimetype="image/png"
        )

    register_error_handler(
        exceptions.TileOutOfBoundsError, handle_tile_out_of_bounds_error
    )

    def handle_dataset_not_found_error(exc: Exception) -> Any:
        # wrong path -> 404
        if current_app.debug:
            raise exc
        return _abort(404, str(exc))

    register_error_handler(
        exceptions.DatasetNotFoundError, handle_dataset_not_found_error
    )

    def handle_database_not_writable_error(exc: Exception) -> Any:
        # database not writable -> 403
        if current_app.debug:
            raise exc
        return _abort(403, str(exc))

    register_error_handler(
        exceptions.DatabaseNotWritableError, handle_database_not_writable_error
    )

    def handle_marshmallow_validation_error(exc: Exception) -> Any:
        # wrong query arguments -> 400
        if current_app.debug:
            raise exc
        return _abort(400, str(exc))

    validation_errors = (
        exceptions.InvalidArgumentsError,
        exceptions.InvalidKeyError,
        marshmallow.ValidationError,
    )

    for err in validation_errors:
        register_error_handler(err, handle_marshmallow_validation_error)


def create_app(debug: bool = False, profile: bool = False) -> Flask:
    """Returns a Flask app"""
    from terracotta import get_settings
    import terracotta.server.datasets
    import terracotta.server.keys
    import terracotta.server.colormap
    import terracotta.server.metadata
    import terracotta.server.rgb
    import terracotta.server.singleband
    import terracotta.server.compute

    new_app = Flask("terracotta.server")
    new_app.debug = debug

    # extensions might modify the global blueprints, so copy before use
    new_tile_api = copy.deepcopy(TILE_API)
    new_metadata_api = copy.deepcopy(METADATA_API)

    # suppress implicit sort of JSON responses
    new_app.config["JSON_SORT_KEYS"] = False

    # CORS
    settings = get_settings()
    CORS(new_tile_api, origins=settings.ALLOWED_ORIGINS_TILES)
    CORS(new_metadata_api, origins=settings.ALLOWED_ORIGINS_METADATA)

    new_app.register_blueprint(new_tile_api, url_prefix="")
    new_app.register_blueprint(new_metadata_api, url_prefix="")

    # register routes on API spec
    with new_app.test_request_context():
        SPEC.path(view=terracotta.server.datasets.get_datasets)
        SPEC.path(view=terracotta.server.keys.get_keys)
        SPEC.path(view=terracotta.server.colormap.get_colormap)
        SPEC.path(view=terracotta.server.metadata.get_metadata)
        SPEC.path(view=terracotta.server.rgb.get_rgb)
        SPEC.path(view=terracotta.server.rgb.get_rgb_preview)
        SPEC.path(view=terracotta.server.singleband.get_singleband)
        SPEC.path(view=terracotta.server.singleband.get_singleband_preview)
        SPEC.path(view=terracotta.server.compute.get_compute)
        SPEC.path(view=terracotta.server.compute.get_compute_preview)

    import terracotta.server.spec

    new_app.register_blueprint(SPEC_API, url_prefix="")

    if profile:
        from werkzeug.contrib.profiler import ProfilerMiddleware

        new_app.wsgi_app = ProfilerMiddleware(new_app.wsgi_app, restrictions=[30])

    _setup_error_handlers(new_app)

    return new_app



import os
from typing import Dict, List

import terracotta

# Define the location of the SQLite database
# (this will be created if it doesn't already exist)
DB_NAME = "./terracotta.sqlite"

# Define the list of keys that will be used to identify datasets.
# (these must match the keys of the "key_values" dicts defined in
# RASTER_FILES)
KEYS = ["type", "rp", "rcp", "epoch", "gcm"]

# Define a list of raster files to import
# (this is a list of dictionaries, each with a file path and the
# values for each key - make sure the order matches the order of
# KEYS defined above)
#
# This part of the script could be replaced with something that
# makes sense for your data - it could use a glob expression to
# find all TIFFs and a regular expression pattern to extract the
# key values, or it could read from a CSV, or use some other
# reference or metadata generating process.
RASTER_FILES = [
    {
        "key_values": {
            "type": "river",
            "rp": 250,
            "rcp": 4.5,
            "epoch": 2030,
            "gcm": "NorESM1-M",
        },
        "path": "./data/river__rp_250__rcp_4x5__epoch_2030__gcm_NorESM1-M.tif",
    },
    {
        "key_values": {
            "type": "river",
            "rp": 500,
            "rcp": 8.5,
            "epoch": 2080,
            "gcm": "NorESM1-M",
        },
        "path": "./data/river__rp_500__rcp_8x5__epoch_2080__gcm_NorESM1-M.tif",
    },
]


def load(db_name: str, keys: List[str], raster_files: List[Dict]):
    # get a TerracottaDriver that we can use to interact with
    # the database
    driver = terracotta.get_driver(db_name)

    # create the database file if it doesn't exist already
    if not os.path.isfile(db_name):
        driver.create(keys)

    # check that the database has the same keys that we want
    # to load
    assert list(driver.key_names) == keys, (driver.key_names, keys)

    # connect to the database
    with driver.connect():
        # insert metadata for each raster into the database
        for raster in raster_files:
            driver.insert(raster["key_values"], raster["path"])


if __name__ == "__main__":
    load(DB_NAME, KEYS, RASTER_FILES)



import pytest

from click.testing import CliRunner


@pytest.fixture(scope="session")
def toml_file(tmpdir_factory):
    content = """
    DEFAULT_TILE_SIZE = [64, 64]
    """
    outfile = tmpdir_factory.mktemp("config").join("tc-config.toml")
    with open(outfile, "w") as f:
        f.write(content)

    return outfile


def test_serve_from_pattern(raster_file):
    from terracotta.scripts import cli

    input_pattern = str(raster_file.dirpath("{name}.tif"))

    runner = CliRunner()
    result = runner.invoke(cli.cli, ["serve", "-r", input_pattern])
    assert result.exit_code == 0


def test_serve_from_database(testdb):
    from terracotta.scripts import cli

    runner = CliRunner()
    result = runner.invoke(cli.cli, ["serve", "-d", str(testdb)])
    assert result.exit_code == 0


def test_serve_no_args(testdb):
    from terracotta.scripts import cli

    runner = CliRunner()
    result = runner.invoke(cli.cli, ["serve"])
    assert result.exit_code != 0


def test_serve_with_config(testdb, toml_file):
    from terracotta.scripts import cli

    runner = CliRunner()
    result = runner.invoke(
        cli.cli, ["--config", str(toml_file), "serve", "-d", str(testdb)]
    )
    assert result.exit_code == 0


def test_serve_rgb_key(raster_file):
    from terracotta.scripts import cli

    input_pattern = str(raster_file.dirpath("{rgb}m{foo}.tif"))

    runner = CliRunner()
    result = runner.invoke(cli.cli, ["serve", "-r", input_pattern, "--rgb-key", "rgb"])
    assert result.exit_code == 0


def test_serve_invalid_rgb_key(raster_file):
    from terracotta.scripts import cli

    input_pattern = str(raster_file.dirpath("{rgb}m{foo}.tif"))

    runner = CliRunner()
    result = runner.invoke(cli.cli, ["serve", "-r", input_pattern, "--rgb-key", "bar"])
    assert result.exit_code != 0


def test_serve_find_socket(raster_file):
    import socket
    from contextlib import closing

    from terracotta.scripts import cli

    input_pattern = str(raster_file.dirpath("{name}.tif"))

    host = "127.0.0.1"
    port = 5000

    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.bind((host, port))

        runner = CliRunner()
        result = runner.invoke(
            cli.cli, ["serve", "-r", input_pattern, "--port", "5000"]
        )
        assert result.exit_code != 0
        assert "Could not find open port to bind to" in result.output

        runner = CliRunner()
        result = runner.invoke(cli.cli, ["serve", "-r", input_pattern])
        assert result.exit_code == 0



"""server/fields.py

Custom marshmallow fields for the server API.
"""

import re
from marshmallow import ValidationError, fields

from typing import Any, Union


class StringOrNumber(fields.Field):
    """
    Marshmallow type that can be either a string or a number.
    Uses marshmallow's default serialization/deserialization
    for `String` or a `Float` depending on the value type.
    """

    def _serialize(
        self, value: Union[str, bytes, int, float], attr: Any, obj: Any, **kwargs: Any
    ) -> Union[str, float, None]:
        if isinstance(value, (str, bytes)):
            return fields.String()._serialize(value, attr, obj, **kwargs)
        elif isinstance(value, (int, float)):
            return fields.Float()._serialize(value, attr, obj, **kwargs)
        else:
            raise ValidationError("Must be a string or a number")

    def _deserialize(
        self, value: Union[str, bytes, int, float], attr: Any, data: Any, **kwargs: Any
    ) -> Union[str, float, None]:
        if isinstance(value, (str, bytes)):
            return fields.String()._deserialize(value, attr, data, **kwargs)
        elif isinstance(value, (int, float)):
            return fields.Float()._deserialize(value, attr, data, **kwargs)
        else:
            raise ValidationError("Must be a string or a number")


def validate_stretch_range(data: Any) -> None:
    """
    Validates that the stretch range is in the format `p<digits>`
    when a string is used.
    """
    if isinstance(data, str):
        if not re.match("^p\\d+$", data):
            raise ValidationError("Percentile format is `p<digits>`")



"""SQLite-remote driver specific tests.

Tests that apply to all drivers go to test_drivers.py.
"""

import os
import tempfile
import time
import uuid
from pathlib import Path

import pytest

from terracotta import exceptions

boto3 = pytest.importorskip("boto3")
moto = pytest.importorskip("moto")


@pytest.fixture(autouse=True)
def mock_aws_env(monkeypatch):
    with monkeypatch.context() as m:
        m.setenv("AWS_DEFAULT_REGION", "us-east-1")
        m.setenv("AWS_ACCESS_KEY_ID", "FakeKey")
        m.setenv("AWS_SECRET_ACCESS_KEY", "FakeSecretKey")
        m.setenv("AWS_SESSION_TOKEN", "FakeSessionToken")
        yield


class Timer:
    def __init__(self, auto=False):
        self.auto = auto
        self.time = 0

    def __call__(self):
        if self.auto:
            self.time += 1
        return self.time

    def tick(self):
        self.time += 1


@pytest.fixture()
def s3_db_factory(tmpdir):
    bucketname = str(uuid.uuid4())

    def _s3_db_factory(keys, datasets=None):
        from terracotta import get_driver

        with tempfile.TemporaryDirectory() as tmpdir:
            dbfile = Path(tmpdir) / "tc.sqlite"
            driver = get_driver(dbfile)
            driver.create(keys)

            with driver.connect():
                if datasets:
                    for keys, path in datasets.items():
                        driver.insert(keys, path)

            # make sure that the connection is closed
            driver.meta_store.sqla_engine.dispose()

            with open(dbfile, "rb") as f:
                db_bytes = f.read()

        conn = boto3.resource("s3")
        conn.create_bucket(Bucket=bucketname)

        s3 = boto3.client("s3")
        s3.put_object(Bucket=bucketname, Key="tc.sqlite", Body=db_bytes)

        return f"s3://{bucketname}/tc.sqlite"

    return _s3_db_factory


@moto.mock_s3
def test_remote_database(s3_db_factory):
    keys = ("some", "keys")
    dbpath = s3_db_factory(keys)

    from terracotta import get_driver

    driver = get_driver(dbpath)

    assert driver.key_names == keys


def test_invalid_url():
    from terracotta import get_driver

    with pytest.raises(ValueError):
        get_driver("foo", provider="sqlite-remote")


@moto.mock_s3
def test_nonexisting_url():
    from terracotta import exceptions, get_driver

    with pytest.raises(exceptions.InvalidDatabaseError):
        get_driver("s3://foo/db.sqlite")


@moto.mock_s3
def test_remote_database_cache(s3_db_factory, raster_file, monkeypatch):
    keys = ("some", "keys")
    dbpath = s3_db_factory(keys)

    from terracotta import get_driver

    driver = get_driver(dbpath)
    driver.meta_store._last_updated = -float("inf")

    with driver.connect():
        assert driver.key_names == keys
        assert driver.get_datasets() == {}
        modification_date = os.path.getmtime(driver.meta_store._local_path)

        s3_db_factory(keys, datasets={("some", "value"): str(raster_file)})

        # no change yet
        assert driver.get_datasets() == {}
        assert os.path.getmtime(driver.meta_store._local_path) == modification_date

    # check if remote db is cached correctly
    driver.meta_store._last_updated = time.time()

    with driver.connect():  # db connection is cached; so still no change
        assert driver.get_datasets() == {}
        assert os.path.getmtime(driver.meta_store._local_path) == modification_date

    # invalidate cache
    driver.meta_store._last_updated = -float("inf")

    with driver.connect():  # now db is updated on reconnect
        assert list(driver.get_datasets().keys()) == [("some", "value")]
        assert os.path.getmtime(driver.meta_store._local_path) != modification_date


@moto.mock_s3
def test_immutability(s3_db_factory, raster_file):
    keys = ("some", "keys")
    dbpath = s3_db_factory(keys, datasets={("some", "value"): str(raster_file)})

    from terracotta import get_driver

    driver = get_driver(dbpath)

    with pytest.raises(exceptions.DatabaseNotWritableError):
        driver.create(keys)

    with pytest.raises(exceptions.DatabaseNotWritableError):
        driver.insert(("some", "value"), str(raster_file))

    with pytest.raises(exceptions.DatabaseNotWritableError):
        driver.delete(("some", "value"))


@moto.mock_s3
def test_destructor(s3_db_factory, raster_file, capsys):
    keys = ("some", "keys")
    dbpath = s3_db_factory(keys, datasets={("some", "value"): str(raster_file)})

    from terracotta import get_driver

    driver = get_driver(dbpath)
    assert os.path.isfile(driver.meta_store._local_path)

    driver.meta_store.__del__()
    assert not os.path.isfile(driver.meta_store._local_path)

    captured = capsys.readouterr()
    assert "Exception ignored" not in captured.err

    # re-create file to prevent actual destructor from failing
    with open(driver.meta_store._local_path, "w"):
        pass



"""expressions.py

Safe execution of user-supplied math expressions
"""

from typing import Mapping, Dict, Tuple, Callable, Type, Any
import ast
import operator
import concurrent.futures

import numpy as np


EXTRA_CALLABLES: Dict[str, Tuple[Callable, int]] = {
    # 'name': (callable, nargs)
    # mask operations
    "where": (np.ma.where, 3),
    "getmask": (np.ma.getmaskarray, 1),
    "setmask": (lambda arr, mask: np.ma.masked_array(arr, mask=mask), 2),
    "masked_equal": (np.ma.masked_equal, 2),
    "masked_greater": (np.ma.masked_greater, 2),
    "masked_greater_equal": (np.ma.masked_greater_equal, 2),
    "masked_inside": (np.ma.masked_inside, 3),
    "masked_invalid": (np.ma.masked_invalid, 1),
    "masked_less": (np.ma.masked_less, 2),
    "masked_less_equal": (np.ma.masked_less_equal, 2),
    "masked_not_equal": (np.ma.masked_not_equal, 2),
    "masked_outside": (np.ma.masked_outside, 3),
    "masked_where": (np.ma.masked_where, 2),
    # math
    "minimum": (np.minimum, 2),
    "maximum": (np.maximum, 2),
    "abs": (np.abs, 1),
    "mod": (np.mod, 1),
    "sqrt": (np.sqrt, 1),
    "log": (np.log, 1),
    "log10": (np.log10, 1),
    "exp": (np.exp, 1),
    "sin": (np.sin, 1),
    "cos": (np.cos, 1),
    "tan": (np.tan, 1),
    "sinh": (np.sinh, 1),
    "cosh": (np.cosh, 1),
    "tanh": (np.tanh, 1),
    "arcsin": (np.arcsin, 1),
    "arccos": (np.arccos, 1),
    "arctan": (np.arctan, 1),
    "arcsinh": (np.arcsinh, 1),
    "arccosh": (np.arccosh, 1),
    "arctanh": (np.arctanh, 1),
}

EXTRA_CONSTANTS = {
    "pi": np.pi,
    "nan": np.nan,
    "inf": np.inf,
    "nomask": np.ma.nomask,
}


class ParseException(Exception):
    pass


class ExpressionParser(ast.NodeVisitor):
    NODE_TO_BINOP: Dict[Type[ast.operator], Callable] = {
        # math
        ast.Add: operator.add,
        ast.Sub: operator.sub,
        ast.Mult: operator.mul,
        ast.Div: operator.truediv,
        ast.Mod: operator.mod,
        ast.Pow: operator.pow,
        # logic
        ast.BitAnd: operator.and_,
        ast.BitOr: operator.or_,
    }

    NODE_TO_UNOP: Dict[Type[ast.unaryop], Callable] = {
        ast.Invert: operator.invert,
        ast.USub: operator.neg,
    }

    NODE_TO_COMPOP: Dict[Type[ast.cmpop], Callable] = {
        ast.Eq: operator.eq,
        ast.NotEq: operator.ne,
        ast.Lt: operator.lt,
        ast.LtE: operator.le,
        ast.Gt: operator.gt,
        ast.GtE: operator.ge,
    }

    def __init__(
        self,
        constants: Mapping[str, Any],
        callables: Mapping[str, Tuple[Callable, int]],
    ) -> None:
        self.constants = constants
        self.callables = callables

    def generic_visit(self, node: ast.AST) -> None:
        # only visit allowed nodes
        raise ParseException(f"{type(node).__name__} not allowed in expressions")

    def visit_Expression(self, node: ast.Expression) -> Any:
        return self.visit(node.body)

    def visit_Name(self, node: ast.Name) -> Any:
        if node.id in self.constants:
            return self.constants[node.id]

        if node.id in self.callables:
            # return (function_name, callable, num_args)
            return (node.id, *self.callables[node.id])

        raise ParseException(f"unrecognized name '{node.id}' in expression")

    def visit_Call(self, node: ast.Call) -> Any:
        funcname, func, nargs = self.visit(node.func)
        got_nargs = len(node.args)
        if got_nargs != nargs:
            raise ParseException(
                f"wrong number of arguments for function {funcname} "
                f"(got {got_nargs}, expected {nargs})"
            )
        return func(*map(self.visit, node.args))

    def visit_Num(self, node: ast.Num) -> Any:
        return node.n

    def visit_UnaryOp(self, node: ast.UnaryOp) -> Any:
        op_type = type(node.op)
        if op_type not in ExpressionParser.NODE_TO_UNOP:
            raise ParseException(
                f"unary operator {op_type.__name__} not allowed in expressions"
            )

        op_callable = ExpressionParser.NODE_TO_UNOP[op_type]
        return op_callable(self.visit(node.operand))

    def visit_BinOp(self, node: ast.BinOp) -> Any:
        op_type = type(node.op)
        if op_type not in ExpressionParser.NODE_TO_BINOP:
            raise ParseException(
                f"binary operator {op_type.__name__} not allowed in expressions"
            )

        op_callable = ExpressionParser.NODE_TO_BINOP[op_type]
        return op_callable(self.visit(node.left), self.visit(node.right))

    def visit_Compare(self, node: ast.Compare) -> Any:
        if len(node.ops) > 1:
            raise ParseException("chained comparisons are not supported")

        op_type = type(node.ops[0])
        if op_type not in ExpressionParser.NODE_TO_COMPOP:
            raise ParseException(
                f"comparison operator {op_type.__name__} not allowed in expressions"
            )

        op_callable = ExpressionParser.NODE_TO_COMPOP[op_type]
        return op_callable(self.visit(node.left), self.visit(node.comparators[0]))


def evaluate_expression(
    expr: str, operands: Mapping[str, np.ndarray], timeout: float = 1.0
) -> np.ndarray:
    try:
        expr_ast = ast.parse(expr, filename="<expression>", mode="eval")
    except SyntaxError as exc:
        raise ValueError(f"given string {expr} is not a valid expression") from exc

    eval_constants = dict(**operands, **EXTRA_CONSTANTS)
    parser = ExpressionParser(eval_constants, EXTRA_CALLABLES)

    with concurrent.futures.ThreadPoolExecutor(1) as executor:
        future = executor.submit(parser.visit, expr_ast)

        try:
            result = future.result(timeout=timeout)

        except concurrent.futures.TimeoutError:
            raise RuntimeError("timeout during pattern evaluation")

        except ParseException as exc:
            raise ValueError(str(exc)) from None

        except Exception as exc:
            # pass only exception message to not leak traceback
            raise ValueError(
                f"unexpected error while evaluating expression: {exc!s}"
            ) from None

    if not isinstance(result, np.ndarray):
        raise ValueError("expression does not return an array")

    # mask inf and nan values
    result = np.ma.masked_invalid(result)

    return result



# shapely has to be imported before rasterio to work around
# https://github.com/Toblerity/Shapely/issues/553
import shapely.geometry  # noqa: F401

import os
import multiprocessing
import time
from functools import partial

import pytest

import numpy as np
import rasterio


def pytest_configure(config):
    os.environ["TC_TESTING"] = "1"


def pytest_unconfigure(config):
    os.environ["TC_TESTING"] = "0"


@pytest.fixture(autouse=True)
def restore_settings():
    """Wipe settings after every test"""
    import terracotta
    from terracotta.config import TerracottaSettings

    try:
        yield
    finally:
        terracotta._settings = TerracottaSettings()
        terracotta._overwritten_settings = set()


def pytest_addoption(parser):
    parser.addoption(
        "--mysql-server",
        help="MySQL server to use for testing in the form of user:password@host:port",
    )
    parser.addoption(
        "--postgresql-server",
        help="PostgreSQL server to use for testing in the form of user:password@host:port",
    )


@pytest.fixture()
def mysql_server(request):
    return request.config.getoption("mysql_server")


@pytest.fixture()
def postgresql_server(request):
    return request.config.getoption("postgresql_server")


def cloud_optimize(raster_file, outfile, create_mask=False, remove_nodata=False):
    import math
    import contextlib
    import rasterio
    import rasterio.io
    import rasterio.shutil

    COG_PROFILE = {
        "count": 1,
        "driver": "GTiff",
        "interleave": "pixel",
        "tiled": True,
        "blockxsize": 256,
        "blockysize": 256,
        "compress": "DEFLATE",
        "photometric": "MINISBLACK",
        "BIGTIFF": "IF_SAFER",
    }

    with contextlib.ExitStack() as es:
        es.enter_context(
            rasterio.Env(
                GDAL_TIFF_INTERNAL_MASK=True,
                GDAL_TIFF_OVR_BLOCKSIZE=256,
            )
        )
        src = es.enter_context(rasterio.open(str(raster_file)))

        profile = src.profile.copy()
        profile.update(COG_PROFILE)

        if remove_nodata:
            profile["nodata"] = None

        memfile = es.enter_context(rasterio.io.MemoryFile())
        dst = es.enter_context(memfile.open(**profile))

        dst.write(src.read())

        if create_mask:
            dst.write_mask(src.dataset_mask().astype("uint8"))

        max_overview_level = math.ceil(
            math.log2(
                max(
                    dst.height // profile["blockysize"],
                    dst.width // profile["blockxsize"],
                )
            )
        )

        overviews = [2**j for j in range(1, max_overview_level + 1)]
        rs_method = rasterio.enums.Resampling.nearest
        dst.build_overviews(overviews, rs_method)
        rasterio.shutil.copy(dst, str(outfile), copy_src_overviews=True, **COG_PROFILE)


@pytest.fixture(scope="session")
def raster_file(tmpdir_factory):
    import affine

    raster_data = np.arange(-128 * 256, 128 * 256, dtype="int16").reshape(256, 256)

    # Sprinkle in some more nodata
    raster_data.flat[::5] = 10000

    profile = {
        "driver": "GTiff",
        "dtype": "int16",
        "nodata": 10000,
        "width": raster_data.shape[1],
        "height": raster_data.shape[0],
        "count": 1,
        "crs": {"init": "epsg:32637"},
        "transform": affine.Affine(2.0, 0.0, 694920.0, 0.0, -2.0, 2055666.0),
    }

    outpath = tmpdir_factory.mktemp("raster")
    unoptimized_raster = outpath.join("img-raw.tif")
    with rasterio.open(str(unoptimized_raster), "w", **profile) as dst:
        dst.write(raster_data, 1)

    optimized_raster = outpath.join("img.tif")
    cloud_optimize(unoptimized_raster, optimized_raster)

    return optimized_raster


@pytest.fixture(scope="session")
def raster_file_float(raster_file, tmpdir_factory):
    with rasterio.open(str(raster_file)) as src:
        raster_data = src.read().astype("float32")

        # Add some NaNs
        raster_data[:, :100, :100] = np.nan

        profile = src.profile.copy()
        profile.update(dtype="float32")

    outpath = tmpdir_factory.mktemp("raster-float")
    unoptimized_raster = outpath.join("img-raw.tif")
    with rasterio.open(str(unoptimized_raster), "w", **profile) as dst:
        dst.write(
            raster_data,
        )

    optimized_raster = outpath.join("img.tif")
    cloud_optimize(unoptimized_raster, optimized_raster)

    return optimized_raster


@pytest.fixture(scope="session")
def big_raster_file_nodata(tmpdir_factory):
    import affine

    np.random.seed(17)
    raster_data = np.random.randint(
        0, np.iinfo(np.uint16).max, size=(2048, 2048), dtype="uint16"
    )
    nodata = 10000

    # include some big nodata regions
    ix, iy = np.indices(raster_data.shape)
    circular_mask = (
        np.sqrt(
            (ix - raster_data.shape[0] / 2) ** 2 + (iy - raster_data.shape[1] / 2) ** 2
        )
        > 1000
    )
    raster_data[circular_mask] = nodata
    raster_data[500:1000, 1000:2000] = nodata
    raster_data[1200, :] = nodata

    profile = {
        "driver": "GTiff",
        "dtype": "uint16",
        "nodata": nodata,
        "width": raster_data.shape[1],
        "height": raster_data.shape[0],
        "count": 1,
        "crs": {"init": "epsg:32637"},
        "transform": affine.Affine(10.0, 0.0, 694920.0, 0.0, -10.0, 2055666.0),
    }

    outpath = tmpdir_factory.mktemp("raster")
    unoptimized_raster = outpath.join("img-raw.tif")
    with rasterio.open(str(unoptimized_raster), "w", **profile) as dst:
        dst.write(raster_data, 1)

    optimized_raster = outpath.join("img-nodata.tif")
    cloud_optimize(unoptimized_raster, optimized_raster)

    return optimized_raster


@pytest.fixture(scope="session")
def big_raster_file_mask(tmpdir_factory, big_raster_file_nodata):
    outpath = tmpdir_factory.mktemp("raster")
    optimized_raster = outpath.join("img-alpha.tif")
    cloud_optimize(
        big_raster_file_nodata, optimized_raster, create_mask=True, remove_nodata=False
    )
    return optimized_raster


@pytest.fixture(scope="session")
def big_raster_file_nomask(tmpdir_factory, big_raster_file_nodata):
    outpath = tmpdir_factory.mktemp("raster")
    optimized_raster = outpath.join("img-alpha.tif")
    cloud_optimize(
        big_raster_file_nodata, optimized_raster, create_mask=False, remove_nodata=True
    )
    return optimized_raster


@pytest.fixture(scope="session")
def unoptimized_raster_file(tmpdir_factory):
    import affine

    np.random.seed(17)
    raster_data = np.random.randint(
        0, np.iinfo(np.uint16).max, size=(1024, 1024), dtype="uint16"
    )
    nodata = 10000

    # include some big nodata regions
    ix, iy = np.indices(raster_data.shape)
    circular_mask = (
        np.sqrt(
            (ix - raster_data.shape[0] / 2) ** 2 + (iy - raster_data.shape[1] / 2) ** 2
        )
        > 400
    )
    raster_data[circular_mask] = nodata
    raster_data[200:600, 400:800] = nodata
    raster_data[500, :] = nodata

    profile = {
        "driver": "GTiff",
        "dtype": "uint16",
        "nodata": nodata,
        "width": raster_data.shape[1],
        "height": raster_data.shape[0],
        "count": 1,
        "crs": {"init": "epsg:32637"},
        "transform": affine.Affine(2.0, 0.0, 694920.0, 0.0, -2.0, 2055666.0),
    }

    outpath = tmpdir_factory.mktemp("raster")
    unoptimized_raster = outpath.join("img-raw.tif")
    with rasterio.open(str(unoptimized_raster), "w", **profile) as dst:
        dst.write(raster_data, 1)

    return unoptimized_raster


@pytest.fixture(scope="session")
def invalid_raster_file(tmpdir_factory):
    """A raster file that is all nodata"""
    import affine

    raster_data = np.full((256, 256), 0, dtype="uint16")
    profile = {
        "driver": "GTiff",
        "dtype": "uint16",
        "nodata": 0,
        "width": raster_data.shape[1],
        "height": raster_data.shape[0],
        "count": 1,
        "crs": {"init": "epsg:32637"},
        "transform": affine.Affine(2.0, 0.0, 694920.0, 0.0, -2.0, 2055666.0),
    }

    outpath = tmpdir_factory.mktemp("raster")
    unoptimized_raster = outpath.join("img-raw.tif")
    with rasterio.open(str(unoptimized_raster), "w", **profile) as dst:
        dst.write(raster_data, 1)

    optimized_raster = outpath.join("img.tif")
    cloud_optimize(unoptimized_raster, optimized_raster)

    return optimized_raster


@pytest.fixture(scope="session")
def raster_file_xyz(raster_file):
    import rasterio
    import rasterio.warp
    import mercantile

    with rasterio.open(str(raster_file)) as src:
        raster_bounds = rasterio.warp.transform_bounds(
            src.crs, "epsg:4326", *src.bounds
        )
    raster_center_x = (raster_bounds[0] + raster_bounds[2]) / 2
    raster_center_y = (raster_bounds[1] + raster_bounds[3]) / 2

    zoom = 14
    tile = mercantile.tile(raster_center_x, raster_center_y, zoom)
    return (tile.x, tile.y, zoom)


@pytest.fixture(scope="session")
def raster_file_xyz_lowzoom(raster_file):
    import rasterio
    import rasterio.warp
    import mercantile

    with rasterio.open(str(raster_file)) as src:
        raster_bounds = rasterio.warp.transform_bounds(
            src.crs, "epsg:4326", *src.bounds
        )
    raster_center_x = (raster_bounds[0] + raster_bounds[2]) / 2
    raster_center_y = (raster_bounds[1] + raster_bounds[3]) / 2

    zoom = 10
    tile = mercantile.tile(raster_center_x, raster_center_y, zoom)
    return (tile.x, tile.y, zoom)


@pytest.fixture(scope="session")
def testdb(raster_file, tmpdir_factory):
    """A read-only, pre-populated test database"""
    from terracotta import get_driver

    keys = ["key1", "akey", "key2"]
    key_descriptions = {"key2": "key2"}

    dbpath = tmpdir_factory.mktemp("db").join("db-readonly.sqlite")
    driver = get_driver(dbpath, provider="sqlite")
    driver.create(keys, key_descriptions=key_descriptions)

    metadata = driver.compute_metadata(str(raster_file), extra_metadata=["extra_data"])

    with driver.connect():
        driver.insert(("val11", "x", "val12"), str(raster_file), metadata=metadata)
        driver.insert(("val21", "x", "val22"), str(raster_file))
        driver.insert(("val21", "x", "val23"), str(raster_file))
        driver.insert(("val21", "x", "val24"), str(raster_file))

    return dbpath


@pytest.fixture()
def v07_db(tmpdir_factory):
    """A read-only, pre-populated test database at terracotta v0.7"""
    import shutil

    dbpath = tmpdir_factory.mktemp("db").join("db-outdated.sqlite")
    shutil.copyfile(
        os.path.join(os.path.dirname(__file__), "data", "db-v075.sqlite"), dbpath
    )

    return dbpath


@pytest.fixture()
def use_testdb(testdb, monkeypatch):
    import terracotta

    terracotta.update_settings(DRIVER_PATH=str(testdb))


@pytest.fixture()
def use_non_writable_testdb(testdb, monkeypatch, raster_file):
    import terracotta

    terracotta.update_settings(DRIVER_PATH=str(testdb))

    driver = terracotta.get_driver(testdb)
    with driver.connect():
        driver.insert(
            ("first", "second", "third"), str(raster_file), skip_metadata=True
        )

    with monkeypatch.context() as m:
        m.setattr(driver.meta_store, "_WRITABLE", False)
        yield

    driver.delete(("first", "second", "third"))


def run_test_server(driver_path, port):
    from terracotta import update_settings

    update_settings(DRIVER_PATH=driver_path)

    from terracotta.server.flask_api import create_app

    create_app().run(port=port)


@pytest.fixture(scope="session")
def test_server(testdb):
    """Spawn a Terracotta server in a separate process"""
    port = 5555
    server_proc = multiprocessing.Process(
        target=partial(run_test_server, driver_path=str(testdb), port=port)
    )
    server_proc.start()
    try:
        # make sure server has started up
        time.sleep(1)
        assert server_proc.is_alive()
        yield f"localhost:{port}"
    finally:
        server_proc.terminate()
        server_proc.join(5)
        assert not server_proc.is_alive()


@pytest.fixture()
def driver_path(provider, tmpdir, mysql_server, postgresql_server):
    """Get a valid, uninitialized driver path for given provider"""
    import random
    import string
    from terracotta import drivers
    from terracotta.exceptions import InvalidDatabaseError
    import sqlalchemy as sqla

    from urllib.parse import urlparse

    def validate_con_info(con_info, db_scheme):
        return (
            con_info.scheme == db_scheme
            and con_info.hostname
            and con_info.username
            and not con_info.path
        )

    def random_string(length):
        return "".join(random.choices(string.ascii_lowercase, k=length))

    if provider == "sqlite":
        dbfile = tmpdir.join("test.sqlite")
        yield str(dbfile)

    elif provider == "mysql" or provider == "postgresql":

        if provider == "mysql":
            db_server = mysql_server
            con_db = "mysql"
        elif provider == "postgresql":
            db_server = postgresql_server
            con_db = "postgres"

        if not db_server:
            return pytest.skip(f"{provider}_server argument not given")

        if not db_server.startswith(f"{provider}://"):
            db_server = f"{provider}://{db_server}"

        con_info = urlparse(db_server)
        if not validate_con_info(con_info, provider):
            raise ValueError(f"invalid value for {provider}_server")

        driver = drivers.get_driver(f"{db_server}/{con_db}", provider=provider)
        try:
            with driver.connect(verify=False):
                pass
        except InvalidDatabaseError as exc:
            raise RuntimeError(f"error connecting to {provider} server") from exc

        dbpath = random_string(24)
        try:
            yield f"{db_server}/{dbpath}"

        finally:  # cleanup
            with driver.meta_store.sqla_engine.connect().execution_options(
                isolation_level="AUTOCOMMIT"
            ) as conn:
                if provider == "postgresql":
                    # Postgres refuses to drop DB if any sessions are hanging around
                    conn.execute(
                        sqla.text(
                            "SELECT pg_terminate_backend(pg_stat_activity.pid) "
                            "FROM pg_stat_activity "
                            f"WHERE pg_stat_activity.datname = '{dbpath}'"
                        )
                    )
                conn.execute(sqla.text(f"DROP DATABASE IF EXISTS {dbpath}"))

    else:
        return NotImplementedError(f"unknown provider {provider}")


@pytest.fixture()
def force_reload():
    """Force a reload of the Terracotta module"""
    import sys

    def purge_module(modname):
        for mod in list(sys.modules.values()):
            if mod is None:
                continue
            if mod.__name__ == modname or mod.__name__.startswith(f"{modname}."):
                del sys.modules[mod.__name__]

    purge_module("terracotta")
    try:
        yield
    finally:
        purge_module("terracotta")



"""drivers/__init__.py

Define an interface to retrieve Terracotta drivers.
"""

import os
from typing import Optional, Union, Tuple, Dict, Type
import urllib.parse as urlparse
from pathlib import Path

from terracotta.drivers.base_classes import MetaStore
from terracotta.drivers.terracotta_driver import TerracottaDriver
from terracotta.drivers.geotiff_raster_store import GeoTiffRasterStore

URLOrPathType = Union[str, Path]


def load_driver(provider: str) -> Type[MetaStore]:
    if provider == "sqlite-remote":
        from terracotta.drivers.sqlite_remote_meta_store import RemoteSQLiteMetaStore

        return RemoteSQLiteMetaStore

    if provider == "mysql":
        from terracotta.drivers.mysql_meta_store import MySQLMetaStore

        return MySQLMetaStore

    if provider == "postgresql":
        from terracotta.drivers.postgresql_meta_store import PostgreSQLMetaStore

        return PostgreSQLMetaStore

    if provider == "sqlite":
        from terracotta.drivers.sqlite_meta_store import SQLiteMetaStore

        return SQLiteMetaStore

    raise ValueError(f"Unknown database provider {provider}")


def auto_detect_provider(url_or_path: str) -> str:
    parsed_path = urlparse.urlparse(url_or_path)

    scheme = parsed_path.scheme
    if scheme == "s3":
        return "sqlite-remote"

    if scheme == "mysql":
        return "mysql"

    if scheme == "postgresql":
        return "postgresql"

    return "sqlite"


_DRIVER_CACHE: Dict[Tuple[URLOrPathType, str, int], TerracottaDriver] = {}


def get_driver(
    url_or_path: URLOrPathType, provider: Optional[str] = None
) -> TerracottaDriver:
    """Retrieve Terracotta driver instance for the given path.

    This function always returns the same instance for identical inputs.

    Warning:

       Always retrieve Driver instances through this function instead of
       instantiating them directly to prevent caching issues.

    Arguments:

        url_or_path: A path identifying the database to connect to.
            The expected format depends on the driver provider.
        provider: Driver provider to use (one of sqlite, sqlite-remote, mysql, postgresql;
            default: auto-detect).

    Example:

        >>> import terracotta as tc
        >>> tc.get_driver('tc.sqlite')
        TerracottaDriver(
            meta_store=SQLiteDriver('/home/terracotta/tc.sqlite'),
            raster_store=GeoTiffRasterStore()
        )
        >>> tc.get_driver('mysql://root@localhost/tc')
        TerracottaDriver(
            meta_store=MySQLDriver('mysql+pymysql://localhost:3306/tc'),
            raster_store=GeoTiffRasterStore()
        )
        >>> # pass provider if path is given in a non-standard way
        >>> tc.get_driver('root@localhost/tc', provider='mysql')
        TerracottaDriver(
            meta_store=MySQLDriver('mysql+pymysql://localhost:3306/tc'),
            raster_store=GeoTiffRasterStore()
        )

    """
    url_or_path = str(url_or_path)

    if provider is None:  # try and auto-detect
        provider = auto_detect_provider(url_or_path)

    DriverClass = load_driver(provider)
    normalized_path = DriverClass._normalize_path(url_or_path)
    cache_key = (normalized_path, provider, os.getpid())

    if cache_key not in _DRIVER_CACHE:
        driver = TerracottaDriver(
            meta_store=DriverClass(url_or_path), raster_store=GeoTiffRasterStore()
        )
        _DRIVER_CACHE[cache_key] = driver

    return _DRIVER_CACHE[cache_key]



"""config.py

Terracotta settings parsing.
"""

from typing import Mapping, Any, Tuple, NamedTuple, Dict, List, Optional
import os
import json
import tempfile
import warnings

from marshmallow import Schema, fields, validate, pre_load, post_load, ValidationError

from terracotta import exceptions


class TerracottaSettings(NamedTuple):
    """Contains all settings for the current Terracotta instance."""

    #: Path to database
    DRIVER_PATH: str = ""

    #: Driver provider to use (sqlite, sqlite-remote, mysql, postgresql; auto-detected by default)
    DRIVER_PROVIDER: Optional[str] = None

    #: Activate debug mode in Flask app
    DEBUG: bool = False

    #: Print profile information after every request
    FLASK_PROFILE: bool = False

    #: Send performance traces to AWS X-Ray
    XRAY_PROFILE: bool = False

    #: Default log level (debug, info, warning, error, critical)
    LOGLEVEL: str = "warning"

    #: Size of raster file in-memory cache in bytes
    RASTER_CACHE_SIZE: int = 1024 * 1024 * 490  # 490 MB

    #: Compression level of raster file in-memory cache, from 0-9
    RASTER_CACHE_COMPRESS_LEVEL: int = 9

    #: Tile size to return if not given in parameters
    DEFAULT_TILE_SIZE: Tuple[int, int] = (256, 256)

    #: Maximum size to use when lazy loading metadata (less is faster but less accurate)
    LAZY_LOADING_MAX_SHAPE: Tuple[int, int] = (1024, 1024)

    #: Compression level of output PNGs, from 0-9
    PNG_COMPRESS_LEVEL: int = 1

    #: Timeout in seconds for database connections
    DB_CONNECTION_TIMEOUT: int = 10

    #: Path where cached remote SQLite databases are stored (when using sqlite-remote provider)
    REMOTE_DB_CACHE_DIR: str = os.path.join(tempfile.gettempdir(), "terracotta")

    #: Time-to-live of remote database cache in seconds
    REMOTE_DB_CACHE_TTL: int = 10 * 60  # 10 min

    #: Resampling method to use when reading reprojected data
    RESAMPLING_METHOD: str = "average"

    #: Resampling method to use when reprojecting data to Web Mercator
    REPROJECTION_METHOD: str = "linear"

    #: CORS allowed origins for metadata endpoint
    ALLOWED_ORIGINS_METADATA: List[str] = ["*"]

    #: CORS allowed origins for tiles endpoints
    ALLOWED_ORIGINS_TILES: List[str] = [r"http[s]?://(localhost|127\.0\.0\.1):*"]

    #: SQL database username (if not given in driver path)
    SQL_USER: Optional[str] = None

    #: SQL database password (if not given in driver path)
    SQL_PASSWORD: Optional[str] = None

    #: Deprecated, use SQL_USER. MySQL database username (if not given in driver path)
    MYSQL_USER: Optional[str] = None

    #: Deprecated, use SQL_PASSWORD. MySQL database password (if not given in driver path)
    MYSQL_PASSWORD: Optional[str] = None

    #: Deprecated, use SQL_USER. PostgreSQL database username (if not given in driver path)
    POSTGRESQL_USER: Optional[str] = None

    #: Deprecated, use SQL_PASSWORD. PostgreSQL database password (if not given in driver path)
    POSTGRESQL_PASSWORD: Optional[str] = None

    #: Use a process pool for band retrieval in parallel
    USE_MULTIPROCESSING: bool = True

    #: Maximum number of metadata keys per POST /metadata request
    MAX_POST_METADATA_KEYS: int = 100


AVAILABLE_SETTINGS: Tuple[str, ...] = TerracottaSettings._fields

DEPRECATION_MAP: Dict[str, str] = {
    # TODO: Remove in v0.8.0
    "MYSQL_USER": "SQL_USER",
    "MYSQL_PASSWORD": "SQL_PASSWORD",
    "POSTGRESQL_USER": "SQL_USER",
    "POSTGRESQL_PASSWORD": "SQL_PASSWORD",
}


def _is_writable(path: str) -> bool:
    return os.access(os.path.dirname(path) or os.getcwd(), os.W_OK)


class SettingSchema(Schema):
    """Schema used to create and validate TerracottaSettings objects"""

    DRIVER_PATH = fields.String()
    DRIVER_PROVIDER = fields.String(allow_none=True)

    DEBUG = fields.Boolean()
    FLASK_PROFILE = fields.Boolean()
    XRAY_PROFILE = fields.Boolean()

    LOGLEVEL = fields.String(
        validate=validate.OneOf(["debug", "info", "warning", "error", "critical"])
    )

    RASTER_CACHE_SIZE = fields.Integer(validate=validate.Range(min=0))
    RASTER_CACHE_COMPRESS_LEVEL = fields.Integer(validate=validate.Range(min=0, max=9))

    DEFAULT_TILE_SIZE = fields.List(fields.Integer(), validate=validate.Length(equal=2))

    LAZY_LOADING_MAX_SHAPE = fields.List(
        fields.Integer(validate=validate.Range(min=0)),
        validate=validate.Length(equal=2),
    )

    PNG_COMPRESS_LEVEL = fields.Integer(validate=validate.Range(min=0, max=9))

    DB_CONNECTION_TIMEOUT = fields.Integer(validate=validate.Range(min=0))
    REMOTE_DB_CACHE_DIR = fields.String(validate=_is_writable)
    REMOTE_DB_CACHE_TTL = fields.Integer(validate=validate.Range(min=0))

    RESAMPLING_METHOD = fields.String(
        validate=validate.OneOf(["nearest", "linear", "cubic", "average"])
    )
    REPROJECTION_METHOD = fields.String(
        validate=validate.OneOf(["nearest", "linear", "cubic", "average"])
    )

    ALLOWED_ORIGINS_METADATA = fields.List(fields.String())
    ALLOWED_ORIGINS_TILES = fields.List(fields.String())

    SQL_USER = fields.String(allow_none=True)
    SQL_PASSWORD = fields.String(allow_none=True)

    MYSQL_USER = fields.String(allow_none=True)
    MYSQL_PASSWORD = fields.String(allow_none=True)
    POSTGRESQL_USER = fields.String(allow_none=True)
    POSTGRESQL_PASSWORD = fields.String(allow_none=True)

    USE_MULTIPROCESSING = fields.Boolean()

    MAX_POST_METADATA_KEYS = fields.Integer(validate=validate.Range(min=1))

    @pre_load
    def decode_lists(self, data: Dict[str, Any], **kwargs: Any) -> Dict[str, Any]:
        for var in (
            "DEFAULT_TILE_SIZE",
            "LAZY_LOADING_MAX_SHAPE",
            "ALLOWED_ORIGINS_METADATA",
            "ALLOWED_ORIGINS_TILES",
        ):
            val = data.get(var)
            if val and isinstance(val, str):
                try:
                    data[var] = json.loads(val)
                except json.decoder.JSONDecodeError as exc:
                    raise ValidationError(
                        f'Could not parse value for key {var} as JSON: "{val}"'
                    ) from exc
        return data

    @pre_load
    def handle_deprecated_fields(
        self, data: Dict[str, Any], **kwargs: Any
    ) -> Dict[str, Any]:
        for deprecated_field, new_field in DEPRECATION_MAP.items():
            if data.get(deprecated_field):
                warnings.warn(
                    f"Setting TC_{deprecated_field} is deprecated "
                    "and will be removed in the next major release. "
                    f"Please use TC_{new_field} instead.",
                    exceptions.DeprecationWarning,
                )

                # Only use the mapping if the new field has not been set
                if not data.get(new_field):
                    data[new_field] = data[deprecated_field]

        return data

    @post_load
    def make_settings(self, data: Dict[str, Any], **kwargs: Any) -> TerracottaSettings:
        # encode tuples
        for var in (
            "DEFAULT_TILE_SIZE",
            "LAZY_LOADING_MAX_SHAPE",
            "ALLOWED_ORIGINS_METADATA",
            "ALLOWED_ORIGINS_TILES",
        ):
            val = data.get(var)
            if val:
                data[var] = tuple(val)

        return TerracottaSettings(**data)


def parse_config(config: Optional[Mapping[str, Any]] = None) -> TerracottaSettings:
    """Parse given config dict and return new TerracottaSettings object"""
    config_dict = dict(config or {})

    for setting in AVAILABLE_SETTINGS:
        env_setting = f"TC_{setting}"
        if setting not in config_dict and env_setting in os.environ:
            config_dict[setting] = os.environ[env_setting]

    schema = SettingSchema()
    try:
        new_settings = schema.load(config_dict)
    except ValidationError as exc:
        raise ValueError("Could not parse configuration") from exc

    return new_settings



"""server/rgb.py

Flask route to handle /rgb calls.
"""

from typing import Optional, Any, Mapping, Dict, Tuple
import json

from marshmallow import Schema, fields, validate, pre_load, ValidationError, EXCLUDE
from flask import request, send_file, Response

from terracotta.server.fields import StringOrNumber, validate_stretch_range
from terracotta.server.flask_api import TILE_API


class RGBQuerySchema(Schema):
    keys = fields.String(
        required=True, description="Keys identifying dataset, in order"
    )
    tile_z = fields.Int(required=True, description="Requested zoom level")
    tile_y = fields.Int(required=True, description="y coordinate")
    tile_x = fields.Int(required=True, description="x coordinate")


class RGBOptionSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    r = fields.String(required=True, description="Key value for red band")
    g = fields.String(required=True, description="Key value for green band")
    b = fields.String(required=True, description="Key value for blue band")
    r_range = fields.List(
        StringOrNumber(allow_none=True, validate=validate_stretch_range),
        validate=validate.Length(equal=2),
        example="[0,1]",
        missing=None,
        description=(
            "Stretch range [min, max] to use for the red band as JSON array. "
            "Min and max may be numbers to use as absolute range, or strings "
            "of the format `p<digits>` with an integer between 0 and 100 "
            "to use percentiles of the image instead. "
            "Null values indicate global minimum / maximum."
        ),
    )
    g_range = fields.List(
        StringOrNumber(allow_none=True, validate=validate_stretch_range),
        validate=validate.Length(equal=2),
        example="[0,1]",
        missing=None,
        description=(
            "Stretch range [min, max] to use for the gren band as JSON array. "
            "Min and max may be numbers to use as absolute range, or strings "
            "of the format `p<digits>` with an integer between 0 and 100 "
            "to use percentiles of the image instead. "
            "Null values indicate global minimum / maximum."
        ),
    )
    b_range = fields.List(
        StringOrNumber(allow_none=True, validate=validate_stretch_range),
        validate=validate.Length(equal=2),
        example="[0,1]",
        missing=None,
        description=(
            "Stretch range [min, max] to use for the blue band as JSON array. "
            "Min and max may be numbers to use as absolute range, or strings "
            "of the format `p<digits>` with an integer between 0 and 100 "
            "to use percentiles of the image instead. "
            "Null values indicate global minimum / maximum."
        ),
    )
    tile_size = fields.List(
        fields.Integer(),
        validate=validate.Length(equal=2),
        example="[256,256]",
        description="Pixel dimensions of the returned PNG image as JSON list.",
    )

    @pre_load
    def process_ranges(self, data: Mapping[str, Any], **kwargs: Any) -> Dict[str, Any]:
        data = dict(data.items())
        for var in ("r_range", "g_range", "b_range", "tile_size"):
            val = data.get(var)
            if val:
                try:
                    data[var] = json.loads(val)
                except json.decoder.JSONDecodeError as exc:
                    raise ValidationError(
                        f"Could not decode value for {var} as JSON"
                    ) from exc
        return data


@TILE_API.route("/rgb/<int:tile_z>/<int:tile_x>/<int:tile_y>.png", methods=["GET"])
@TILE_API.route(
    "/rgb/<path:keys>/<int:tile_z>/<int:tile_x>/<int:tile_y>.png", methods=["GET"]
)
def get_rgb(tile_z: int, tile_y: int, tile_x: int, keys: str = "") -> Response:
    """Return the requested RGB tile as a PNG image.
    ---
    get:
        summary: /rgb (tile)
        description: Combine three datasets to RGB image, and return tile as PNG
        parameters:
            - in: path
              schema: RGBQuerySchema
            - in: query
              schema: RGBOptionSchema
        responses:
            200:
                description:
                    PNG image of requested tile
            400:
                description:
                    Invalid query parameters
            404:
                description:
                    No dataset found for given key combination
    """
    tile_xyz = (tile_x, tile_y, tile_z)
    return _get_rgb_image(keys, tile_xyz=tile_xyz)


class RGBPreviewQuerySchema(Schema):
    keys = fields.String(
        required=True, description="Keys identifying dataset, in order"
    )


@TILE_API.route("/rgb/preview.png", methods=["GET"])
@TILE_API.route("/rgb/<path:keys>/preview.png", methods=["GET"])
def get_rgb_preview(keys: str = "") -> Response:
    """Return the requested RGB dataset preview as a PNG image.
    ---
    get:
        summary: /rgb (preview)
        description: Combine three datasets to RGB image, and return preview as PNG
        parameters:
            - in: path
              schema: RGBPreviewQuerySchema
            - in: query
              schema: RGBOptionSchema
        responses:
            200:
                description:
                    PNG image of requested tile
            400:
                description:
                    Invalid query parameters
            404:
                description:
                    No dataset found for given key combination
    """
    return _get_rgb_image(keys)


def _get_rgb_image(
    keys: str, tile_xyz: Optional[Tuple[int, int, int]] = None
) -> Response:
    from terracotta.handlers.rgb import rgb

    option_schema = RGBOptionSchema()
    options = option_schema.load(request.args)

    some_keys = [key for key in keys.split("/") if key]

    rgb_values = (options.pop("r"), options.pop("g"), options.pop("b"))
    stretch_ranges = tuple(options.pop(k) for k in ("r_range", "g_range", "b_range"))

    image = rgb(
        some_keys,
        rgb_values,
        stretch_ranges=stretch_ranges,
        tile_xyz=tile_xyz,
        **options,
    )

    return send_file(image, mimetype="image/png")



import json

import pytest


def get_client(metadata_origins=None, tile_origins=None):
    from terracotta.server import create_app
    import terracotta

    if metadata_origins is not None:
        terracotta.update_settings(ALLOWED_ORIGINS_METADATA=metadata_origins)

    if tile_origins is not None:
        terracotta.update_settings(ALLOWED_ORIGINS_TILES=tile_origins)

    flask_app = create_app()
    return flask_app.test_client()


@pytest.fixture
def valid_metadata_path():
    return "/metadata/val11/x/val12/"


@pytest.fixture
def valid_singleband_path(raster_file_xyz):
    x, y, z = raster_file_xyz
    return f"/singleband/val11/x/val12/{z}/{x}/{y}.png"


@pytest.mark.parametrize("tile_origins", (None, '["*"]', "[]", '["example.org"]'))
@pytest.mark.parametrize("metadata_origins", (None, '["*"]', "[]", '["example.org"]'))
def test_cors(
    use_testdb,
    valid_metadata_path,
    valid_singleband_path,
    metadata_origins,
    tile_origins,
):
    with get_client(metadata_origins, tile_origins) as client:
        # metadata
        rv = client.get(valid_metadata_path)
        assert rv.status_code == 200

        if metadata_origins == "[]":
            assert "Access-Control-Allow-Origin" not in rv.headers
        elif metadata_origins is None:
            # default for metadata is allow all
            assert rv.headers["Access-Control-Allow-Origin"] == "*"
        else:
            expected = json.loads(metadata_origins)
            if len(expected) == 1:
                expected = expected[0]
            assert rv.headers["Access-Control-Allow-Origin"] == expected

        # tiles
        rv = client.get(valid_singleband_path)
        assert rv.status_code == 200

        if tile_origins == "[]":
            assert "Access-Control-Allow-Origin" not in rv.headers
        elif tile_origins is None:
            # default for tiles is disallow all
            assert "Access-Control-Allow-Origin" not in rv.headers
        else:
            expected = json.loads(tile_origins)
            if len(expected) == 1:
                expected = expected[0]
            assert rv.headers["Access-Control-Allow-Origin"] == expected



"""server/app.py

Instantiated version of the Flask API.
"""

from terracotta import get_settings, logs
from terracotta.server import create_app

settings = get_settings()

logs.set_logger(settings.LOGLEVEL, catch_warnings=True)

app = create_app(debug=settings.DEBUG, profile=settings.FLASK_PROFILE)



"""drivers/sqlite_meta_store.py

SQLite-backed metadata driver. Metadata is stored in an SQLite database.
"""

import os
from pathlib import Path
from typing import Union

from terracotta.drivers.relational_meta_store import RelationalMetaStore


class SQLiteMetaStore(RelationalMetaStore):
    """An SQLite-backed metadata driver.

    Stores metadata and paths to raster files in SQLite.

    This is the simplest Terracotta driver, as it requires no additional infrastructure.
    The SQLite database is simply a file that can e.g. be stored together with the actual
    raster files.

    Note:

        This driver requires the SQLite database to be physically present on the server.
        For remote SQLite databases hosted on S3, use
        :class:`~terracotta.drivers.sqlite_remote.RemoteSQLiteDriver`.

    The SQLite database consists of 4 different tables:

    - ``terracotta``: Metadata about the database itself.
    - ``key_names``: Contains two columns holding all available keys and their description.
    - ``datasets``: Maps key values to physical raster path.
    - ``metadata``: Contains actual metadata as separate columns. Indexed via key values.

    This driver caches key names, but not metadata.

    Warning:

        This driver is not thread-safe. It is not possible to connect to the database
        outside the main thread.

    """

    SQL_DIALECT = "sqlite"
    SQL_DRIVER = "pysqlite"
    SQL_KEY_SIZE = 256
    SQL_TIMEOUT_KEY = "timeout"

    def __init__(self, path: Union[str, Path]) -> None:
        """Initialize the SQLiteDriver.

        This should not be called directly, use :func:`~terracotta.get_driver` instead.

        Arguments:

            path: File path to target SQLite database (may or may not exist yet)

        """
        super().__init__(f"{self.SQL_DIALECT}:///{path}")

    @classmethod
    def _normalize_path(cls, path: str) -> str:
        if path.startswith(f"{cls.SQL_DIALECT}:///"):
            path = path.replace(f"{cls.SQL_DIALECT}:///", "")

        return os.path.normpath(os.path.realpath(path))

    def _create_database(self) -> None:
        """The database is automatically created by the sqlite driver on connection,
        so no need to do anything here
        """
        pass



import json

import pytest


def test_schema_integrity():
    from terracotta import config

    settings_fields = config.TerracottaSettings._fields
    schema_fields = config.SettingSchema._declared_fields
    assert set(settings_fields) == set(schema_fields)


def test_env_config(monkeypatch):
    from terracotta import config

    with monkeypatch.context() as m:
        m.setenv("TC_DRIVER_PATH", "test")
        assert config.parse_config().DRIVER_PATH == "test"

    with monkeypatch.context() as m:
        m.setenv("TC_DRIVER_PATH", "test2")
        assert config.parse_config().DRIVER_PATH == "test2"

    with monkeypatch.context() as m:
        m.setenv("TC_DEFAULT_TILE_SIZE", json.dumps([1, 2]))
        assert config.parse_config().DEFAULT_TILE_SIZE == (1, 2)


def test_env_config_invalid(monkeypatch):
    from terracotta import config

    with monkeypatch.context() as m:
        m.setenv("TC_DEFAULT_TILE_SIZE", "[1")  # unbalanced bracket
        with pytest.raises(ValueError):
            config.parse_config()

    with monkeypatch.context() as m:
        m.setenv("TC_DEBUG", "foo")  # not a boolean
        with pytest.raises(ValueError):
            config.parse_config()

    with monkeypatch.context() as m:
        m.setenv("TC_REMOTE_DB_CACHE_DIR", "/foo/test.sqlite")  # non-existing folder
        with pytest.raises(ValueError):
            config.parse_config()

    assert True


def test_dict_config():
    from terracotta import config

    settings = config.parse_config({"DRIVER_PATH": "test3"})
    assert settings.DRIVER_PATH == "test3"

    settings = config.parse_config({"DEFAULT_TILE_SIZE": [100, 100]})
    assert settings.DEFAULT_TILE_SIZE == (100, 100)


def test_terracotta_settings():
    from terracotta import config

    settings = config.parse_config()

    assert settings.DEFAULT_TILE_SIZE

    with pytest.raises(AttributeError):
        settings.DEFAULT_TILE_SIZE = (10, 10)


def test_update_config():
    from terracotta import get_settings, update_settings

    update_settings(DRIVER_PATH="test")
    new_settings = get_settings()
    assert new_settings.DRIVER_PATH == "test"

    update_settings(DEFAULT_TILE_SIZE=[50, 50])
    new_settings = get_settings()
    assert new_settings.DRIVER_PATH == "test" and new_settings.DEFAULT_TILE_SIZE == (
        50,
        50,
    )


def test_deprecation_behaviour(monkeypatch):
    from terracotta import config, exceptions, get_settings, update_settings

    for deprecated_field, new_field in config.DEPRECATION_MAP.items():
        with monkeypatch.context() as m:
            m.setenv(f"TC_{deprecated_field}", "foo")

            with pytest.warns(exceptions.DeprecationWarning) as warning:
                update_settings()
            assert f"TC_{deprecated_field} is deprecated" in str(warning[0])

            assert getattr(get_settings(), deprecated_field) == "foo"
            assert getattr(get_settings(), new_field) == "foo"

            m.setenv(f"TC_{new_field}", "bar")

            with pytest.warns(exceptions.DeprecationWarning) as warning:
                update_settings()
            assert f"TC_{deprecated_field} is deprecated" in str(warning[0])

            assert getattr(get_settings(), deprecated_field) == "foo"
            assert getattr(get_settings(), new_field) == "bar"



"""drivers/postgresql_meta_store.py

PostgreSQL-backed metadata driver. Metadata is stored in a PostgreSQL database.
"""

from typing import Optional, Mapping, Sequence

import sqlalchemy as sqla
from terracotta.drivers.relational_meta_store import RelationalMetaStore


class PostgreSQLMetaStore(RelationalMetaStore):
    """A PostgreSQL-backed metadata driver.

    Stores metadata and paths to raster files in PostgreSQL.

    Requires a running PostgreSQL server.

    The PostgreSQL database consists of 4 different tables:

    - ``terracotta``: Metadata about the database itself.
    - ``key_names``: Contains two columns holding all available keys and their description.
    - ``datasets``: Maps key values to physical raster path.
    - ``metadata``: Contains actual metadata as separate columns. Indexed via key values.

    This driver caches key names.
    """

    SQL_DIALECT = "postgresql"
    SQL_DRIVER = "psycopg2"
    SQL_TIMEOUT_KEY = "connect_timeout"

    MAX_PRIMARY_KEY_SIZE = 2730 // 4  # Max B-tree index size in bytes
    DEFAULT_PORT = 5432
    # Will connect to this db before creatting the 'terracotta' db
    DEFAULT_CONNECT_DB = "postgres"

    def __init__(self, postgresql_path: str) -> None:
        """Initialize the PostgreSQLDriver.

        This should not be called directly, use :func:`~terracotta.get_driver` instead.

        Arguments:

            postgresql_path: URL to running PostgreSQL server, in the form
                ``postgresql://username:password@hostname/database``

        """
        super().__init__(postgresql_path)

        # raise an exception if database name is invalid
        if not self.url.database:
            raise ValueError("database must be specified in PostgreSQL path")
        if "/" in self.url.database.strip("/"):
            raise ValueError("invalid database path")

    @classmethod
    def _normalize_path(cls, path: str) -> str:
        url = cls._parse_path(path)

        path = f"{url.drivername}://{url.host}:{url.port or cls.DEFAULT_PORT}/{url.database}"
        path = path.rstrip("/")
        return path

    def _create_database(self) -> None:
        engine = sqla.create_engine(
            # `.set()` returns a copy with changed parameters
            self.url.set(database=self.DEFAULT_CONNECT_DB),
            echo=False,
            future=True,
            isolation_level="AUTOCOMMIT",
        )
        with engine.connect() as connection:
            connection.execute(sqla.text(f"CREATE DATABASE {self.url.database}"))
            connection.commit()

    def _initialize_database(
        self, keys: Sequence[str], key_descriptions: Optional[Mapping[str, str]] = None
    ) -> None:
        # Enforce max primary key length equal to max B-tree index size
        self.SQL_KEY_SIZE = self.MAX_PRIMARY_KEY_SIZE // len(keys)
        super()._initialize_database(keys, key_descriptions)



import pytest
from click.testing import CliRunner


def parse_version(verstr):
    """Convert 'v<major>.<minor>.<patch>' to (major, minor, patch)"""
    components = verstr.split(".")
    components[0] = components[0].lstrip("v")
    return tuple(int(c) for c in components[:3])


def migration_testfunc(v07_db, raster_file):
    import terracotta
    from terracotta import get_driver
    from terracotta.scripts import cli

    current_version = parse_version(terracotta.__version__)

    # run migration
    runner = CliRunner()
    result = runner.invoke(
        cli.cli,
        [
            "migrate",
            str(v07_db),
            "--from",
            "v0.7",
            "--to",
            f"v{current_version[0]}.{current_version[1]}",
            "--yes",
        ],
    )

    assert result.exit_code == 0
    assert "Upgrade path found" in result.output

    driver_updated = get_driver(str(v07_db), provider="sqlite")

    # key_names did not exist in v0.7
    assert driver_updated.key_names == ("key1", "akey", "key2")

    # make sure we can still read the data
    datasets = driver_updated.get_datasets()
    assert len(datasets) > 0

    # make sure we can still read metadata
    metadata = driver_updated.get_metadata(next(iter(datasets)))
    assert metadata is not None

    # make sure we can still insert data
    new_dataset_keys = ("1foo", "2bar", "3baz")
    driver_updated.insert(new_dataset_keys, str(raster_file))
    assert new_dataset_keys in driver_updated.get_datasets()


def test_migrate(v07_db, raster_file):
    """Test database migration to this major version."""
    migration_testfunc(v07_db, raster_file)


def test_migrate_next(v07_db, raster_file, monkeypatch, force_reload):
    """Test database migration to next major version if one is available."""
    with monkeypatch.context() as m:
        # pretend we are at the next major version
        import terracotta

        current_version = parse_version(terracotta.__version__)
        next_major_version = (current_version[0], current_version[1] + 1, 0)
        m.setattr(terracotta, "__version__", ".".join(map(str, next_major_version)))

        from terracotta.migrations import MIGRATIONS

        if next_major_version[:2] not in [m.up_version for m in MIGRATIONS.values()]:
            pytest.skip("No migration available for next major version")

        migration_testfunc(v07_db, raster_file)



import logging


def test_logstream(caplog, capsys):
    from terracotta import logs

    assert logs.use_colors

    logger = logs.set_logger("DEBUG")
    caplog.set_level("DEBUG", logger="terracotta")
    logger.warning("test")

    log_out = caplog.records
    assert len(log_out) == 1
    log_out = log_out[0]
    assert "test" == log_out.message
    assert logs.LEVEL_PREFIX["WARNING"] == log_out.levelshortname

    captured = capsys.readouterr()
    assert "[!]" in captured.err
    assert "test" in captured.err


def test_logstream_nocolors(monkeypatch, caplog, capsys):
    with monkeypatch.context() as m:
        from terracotta import logs

        m.setattr(logs, "use_colors", False)

        assert not logs.use_colors

        logger = logs.set_logger("DEBUG")
        caplog.set_level("DEBUG", logger="terracotta")
        logger.warning("test")

        log_out = caplog.records
        assert len(log_out) == 1
        log_out = log_out[0]
        assert "test" == log_out.message
        assert logs.LEVEL_PREFIX["WARNING"] == log_out.levelshortname

        captured = capsys.readouterr()
        assert " [!] test\n" == captured.err


def test_double_init(caplog):
    from terracotta import logs

    logs.set_logger("DEBUG")
    logs.set_logger("DEBUG")

    caplog.set_level("DEBUG", logger="terracotta")
    logger = logging.getLogger("terracotta")
    logger.warning("test")

    assert len(caplog.records) == 1



import numpy as np
from PIL import Image

import pytest


def test_colormap_handler():
    from terracotta.handlers import colormap

    cmap = colormap.colormap(colormap="jet", stretch_range=[0.0, 1.0], num_values=50)
    assert cmap
    assert len(cmap) == 50
    assert len(cmap[0]["rgba"]) == 4
    assert cmap[0]["value"] == 0.0 and cmap[-1]["value"] == 1.0


@pytest.mark.parametrize("stretch_range", [[0, 20000], [20000, 30000], [-50000, 50000]])
@pytest.mark.parametrize("cmap_name", [None, "jet"])
def test_colormap_consistency(
    use_testdb, testdb, raster_file_xyz, stretch_range, cmap_name
):
    """Test consistency between /colormap and images returned by /singleband"""
    import terracotta
    from terracotta.xyz import get_tile_data
    from terracotta.handlers import singleband, colormap

    ds_keys = ["val21", "x", "val22"]

    # get image with applied stretch and colormap
    raw_img = singleband.singleband(
        ds_keys, raster_file_xyz, stretch_range=stretch_range, colormap=cmap_name
    )
    img_data = np.asarray(Image.open(raw_img).convert("RGBA"))

    # get raw data to compare to
    driver = terracotta.get_driver(testdb)

    with driver.connect():
        tile_data = get_tile_data(
            driver, ds_keys, tile_xyz=raster_file_xyz, tile_size=img_data.shape[:2]
        )

    # make sure all pixel values are included in colormap
    num_values = stretch_range[1] - stretch_range[0] + 1

    # get colormap for given stretch
    cmap = colormap.colormap(
        colormap=cmap_name, stretch_range=stretch_range, num_values=num_values
    )
    cmap = dict(row.values() for row in cmap)

    # test nodata
    nodata_mask = tile_data.mask
    assert np.all(img_data[nodata_mask, -1] == 0)

    # test clipping
    below_mask = tile_data < stretch_range[0]
    assert np.all(
        img_data[below_mask & ~nodata_mask, :-1] == cmap[stretch_range[0]][:-1]
    )

    above_mask = tile_data > stretch_range[1]
    assert np.all(
        img_data[above_mask & ~nodata_mask, :-1] == cmap[stretch_range[1]][:-1]
    )

    # test values inside stretch_range
    values_to_test = np.unique(tile_data.compressed())
    values_to_test = values_to_test[
        (values_to_test >= stretch_range[0]) & (values_to_test <= stretch_range[1])
    ]

    for val in values_to_test:
        rgba = cmap[val]
        assert np.all(img_data[tile_data == val] == rgba)


@pytest.mark.parametrize("stretch_range", [[0, 0.1], [20000, 30000], [-50000, 50000]])
def test_colormap_range(stretch_range):
    """Ensure that all colors from colormap file are present in returned data"""
    from terracotta.cmaps import get_cmap
    from terracotta.handlers import colormap

    cmap_name = "jet"
    cmap_in = get_cmap(cmap_name)
    num_values = cmap_in.shape[0]
    cmap_out = colormap.colormap(
        colormap=cmap_name, stretch_range=stretch_range, num_values=num_values
    )
    cmap_out_arr = np.array([val["rgba"] for val in cmap_out], dtype=cmap_in.dtype)
    np.testing.assert_array_equal(cmap_in, cmap_out_arr)



"""drivers/sqlite_remote_meta_store.py

SQLite-backed metadata driver. Metadata is stored in an SQLite database.
"""

import contextlib
import logging
import os
import shutil
import tempfile
import time
import urllib.parse as urlparse
from pathlib import Path
from typing import Iterator, Union

from terracotta import exceptions, get_settings
from terracotta.drivers.sqlite_meta_store import SQLiteMetaStore
from terracotta.profile import trace

logger = logging.getLogger(__name__)


@contextlib.contextmanager
def convert_exceptions(msg: str) -> Iterator:
    """Convert internal boto exceptions to our InvalidDatabaseError"""
    import botocore.exceptions

    try:
        yield
    except botocore.exceptions.ClientError as exc:
        raise exceptions.InvalidDatabaseError(msg) from exc


def _update_from_s3(remote_path: str, local_path: str) -> None:
    import boto3

    parsed_remote_path = urlparse.urlparse(remote_path)
    bucket_name, key = parsed_remote_path.netloc, parsed_remote_path.path.strip("/")

    if parsed_remote_path.scheme != "s3":
        raise ValueError("Expected s3:// URL")

    s3 = boto3.resource("s3")
    obj = s3.Object(bucket_name, key)
    obj_bytes = obj.get()["Body"]

    # copy over existing database; this is somewhat safe since it is read-only
    with open(local_path, "wb") as f:
        shutil.copyfileobj(obj_bytes, f)


class RemoteSQLiteMetaStore(SQLiteMetaStore):
    """An SQLite-backed metadata driver, where the database file is stored remotely on S3.

    Stores metadata and paths to raster files in SQLite.

    See also:

        :class:`~terracotta.drivers.sqlite.SQLiteDriver` for the local version of this
        driver.

    The SQLite database is simply a file that can be stored e.g. together with the actual
    raster files on S3. Before handling the first request, this driver will download a
    temporary copy of the remote database file. It is thus not feasible for large databases.

    The local database copy will be updated in regular intervals defined by
    :attr:`~terracotta.config.TerracottaSettings.REMOTE_DB_CACHE_TTL`.

    Warning:

        This driver is read-only. Any attempts to use the create, insert, or delete methods
        will throw a DatabaseNotWritableError.

    """

    _WRITABLE: bool = False

    def __init__(self, remote_path: Union[str, Path]) -> None:
        """Initialize the RemoteSQLiteDriver.

        This should not be called directly, use :func:`~terracotta.get_driver` instead.

        Arguments:

            remote_path: S3 URL in the form ``s3://bucket/key`` to remote SQLite database
                (has to exist).

        """
        settings = get_settings()

        self.__rm = os.remove  # keep reference to use in __del__

        os.makedirs(settings.REMOTE_DB_CACHE_DIR, exist_ok=True)
        local_db_file = tempfile.NamedTemporaryFile(
            dir=settings.REMOTE_DB_CACHE_DIR,
            prefix="tc_s3_db_",
            suffix=".sqlite",
            delete=False,
        )
        local_db_file.close()

        self._local_path = local_db_file.name
        self._remote_path = str(remote_path)

        # download database
        with convert_exceptions("Could not retrieve database from S3"):
            _update_from_s3(self._remote_path, self._local_path)

        self._last_updated = time.time()

        super().__init__(local_db_file.name)

    @classmethod
    def _normalize_path(cls, path: str) -> str:
        parts = urlparse.urlparse(path)

        if not parts.hostname:
            parts = urlparse.urlparse(f"https://{path}")

        port = parts.port
        if port is None:
            port = 443 if parts.scheme == "https" else 80

        path = f"{parts.scheme}://{parts.hostname}:{port}{parts.path}"
        path = path.rstrip("/")
        return path

    @convert_exceptions("Could not retrieve database from S3")
    @trace("download_db_from_s3")
    def _update_db(self, remote_path: str, local_path: str) -> None:
        settings = get_settings()

        if self._last_updated < time.time() - settings.REMOTE_DB_CACHE_TTL:
            logger.debug("Remote database cache expired, re-downloading")
            _update_from_s3(remote_path, local_path)
            self._last_updated = time.time()

    def _connection_callback(self) -> None:
        self._update_db(self._remote_path, self._local_path)
        super()._connection_callback()

    def __del__(self) -> None:
        """Clean up temporary database upon exit"""
        try:
            self.__rm(self._local_path)
        except OSError:  # pragma: no cover
            pass



import pytest

TEST_CASES = {
    "{provider}://root@localhost:5000/test": dict(
        username="root", password=None, host="localhost", port=5000, database="test"
    ),
    "root@localhost:5000/test": dict(
        username="root", password=None, host="localhost", port=5000, database="test"
    ),
    "{provider}://root:foo@localhost/test": dict(
        username="root", password="foo", host="localhost", port=None, database="test"
    ),
    "{provider}://localhost/test": dict(
        password=None, host="localhost", port=None, database="test"
    ),
    "localhost/test": dict(password=None, host="localhost", port=None, database="test"),
}

INVALID_TEST_CASES = [
    "http://localhost/test",  # wrong scheme
    "{provider}://localhost",  # no database
    "{provider}://localhost/test/foo/bar",  # path too deep
]


@pytest.mark.parametrize("case", TEST_CASES.keys())
@pytest.mark.parametrize("provider", ("mysql", "postgresql"))
def test_path_parsing(case, provider):
    from terracotta import drivers

    # empty cache
    drivers._DRIVER_CACHE = {}

    db_path = case.format(provider=provider)
    db = drivers.get_driver(db_path, provider=provider)
    for attr in ("username", "password", "host", "port", "database"):
        assert getattr(db.meta_store.url, attr) == TEST_CASES[case].get(attr, None)


@pytest.mark.parametrize("case", INVALID_TEST_CASES)
@pytest.mark.parametrize("provider", ("mysql", "postgresql"))
def test_invalid_paths(case, provider):
    from terracotta import drivers

    db_path = case.format(provider=provider)
    with pytest.raises(ValueError):
        drivers.get_driver(db_path, provider=provider)



from io import BytesIO
import json
import urllib.parse

from PIL import Image
import marshmallow
import numpy as np

import pytest


@pytest.fixture(scope="module")
def flask_app():
    from terracotta.server import create_app

    return create_app()


@pytest.fixture(scope="module")
def client(flask_app):
    with flask_app.test_client() as client:
        yield client


@pytest.fixture(scope="module")
def debug_flask_app():
    from terracotta.server import create_app

    return create_app(debug=True)


@pytest.fixture(scope="module")
def debug_client(debug_flask_app):
    with debug_flask_app.test_client() as client:
        yield client


def test_get_keys(client, use_testdb):
    rv = client.get("/keys")

    expected_response = [
        {"key": "key1"},
        {"key": "akey"},
        {"key": "key2", "description": "key2"},
    ]
    assert rv.status_code == 200
    assert expected_response == json.loads(rv.data)["keys"]


def test_get_metadata(client, use_testdb):
    rv = client.get("/metadata/val11/x/val12/")
    assert rv.status_code == 200
    assert ["extra_data"] == json.loads(rv.data)["metadata"]


def test_get_metadata_lazily_nonwritable_db(client, use_non_writable_testdb):
    rv = client.get("/metadata/first/second/third")
    assert rv.status_code == 403


def test_debug_errors(debug_client, use_non_writable_testdb, raster_file_xyz):
    from terracotta import exceptions
    import marshmallow

    with pytest.raises(exceptions.DatabaseNotWritableError):
        debug_client.get("/metadata/first/second/third")

    with pytest.raises(exceptions.DatasetNotFoundError):
        debug_client.get("/metadata/NONEXISTING/KEYS/YO")

    with pytest.raises(exceptions.InvalidKeyError):
        debug_client.get("/metadata/ONLYONEKEY")

    x, y, z = raster_file_xyz

    with pytest.raises(exceptions.InvalidArgumentsError):
        debug_client.get(
            f"/compute/first/second/third/{z}/{x}/{y}.png"
            "?expression=v1*v2&v1=val22&v2=val23"
            "&stretch_range=[10000,0]"
        )

    with pytest.raises(marshmallow.ValidationError):
        debug_client.get(
            f"/compute/first/second/third/{z}/{x}/{y}.png" "?stretch_range=[10000,0]"
        )


def test_get_metadata_nonexisting(client, use_testdb):
    rv = client.get("/metadata/val11/x/NONEXISTING/")
    assert rv.status_code == 404


def test_post_metadata(client, use_testdb):
    rv = client.post(
        "/metadata",
        json={"keys": [["val11", "x", "val12"], ["val21", "x", "val22"]]},
    )

    assert rv.status_code == 200
    assert len(json.loads(rv.data)) == 2


def test_post_metadata_specific_columns(client, use_testdb):
    rv = client.post(
        '/metadata?columns=["bounds", "range"]',
        json={"keys": [["val11", "x", "val12"], ["val21", "x", "val22"]]},
    )

    assert rv.status_code == 200
    assert len(json.loads(rv.data)) == 2
    assert set(json.loads(rv.data)[0].keys()) == {"bounds", "range", "keys"}


def test_post_metadata_errors(debug_client, use_non_writable_testdb):
    from terracotta import exceptions
    import marshmallow

    with pytest.raises(marshmallow.ValidationError):
        debug_client.post(
            '/metadata?columns=["range]',
            json={"keys": [["val11", "x", "val12"], ["val21", "x", "val22"]]},
        )

    with pytest.raises(exceptions.InvalidArgumentsError):
        debug_client.post(
            '/metadata?columns=["range"]',
            json={"keys": [["val11", "x", "val12"] for _ in range(101)]},
        )

    with pytest.raises(exceptions.InvalidArgumentsError):
        debug_client.post(
            '/metadata?columns=["range"]',
            json="Invalid JSON",
        )

    with pytest.raises(KeyError):
        debug_client.post(
            '/metadata?columns=["invalid"]',
            json={"keys": [["val11", "x", "val12"], ["val21", "x", "val22"]]},
        )


def test_get_datasets(client, use_testdb):
    rv = client.get("/datasets")
    assert rv.status_code == 200
    datasets = json.loads(rv.data)["datasets"]
    assert len(datasets) == 4
    assert dict([("key1", "val11"), ("akey", "x"), ("key2", "val12")]) in datasets


def test_get_datasets_pagination(client, use_testdb):
    # no page (implicit 0)
    rv = client.get("/datasets?limit=2")
    assert rv.status_code == 200
    response = json.loads(rv.data)
    assert response["limit"] == 2
    assert response["page"] == 0

    first_datasets = response["datasets"]
    assert len(first_datasets) == 2
    assert dict([("key1", "val11"), ("akey", "x"), ("key2", "val12")]) in first_datasets

    # second page
    rv = client.get("/datasets?limit=2&page=1")
    assert rv.status_code == 200
    response = json.loads(rv.data)
    assert response["limit"] == 2
    assert response["page"] == 1

    last_datasets = response["datasets"]
    assert len(last_datasets) == 2
    assert (
        dict([("key1", "val11"), ("akey", "x"), ("key2", "val12")]) not in last_datasets
    )

    # page out of range
    rv = client.get("/datasets?limit=2&page=1000")
    assert rv.status_code == 200
    assert not json.loads(rv.data)["datasets"]

    # invalid page
    rv = client.get("/datasets?page=-1")
    assert rv.status_code == 400

    # invalid limit
    rv = client.get("/datasets?limit=-1")
    assert rv.status_code == 400


def test_get_datasets_selective(client, use_testdb):
    rv = client.get("/datasets?key1=val21")
    assert rv.status_code == 200
    assert len(json.loads(rv.data)["datasets"]) == 3

    rv = client.get("/datasets?key1=val21&key2=val23")
    assert rv.status_code == 200
    assert len(json.loads(rv.data)["datasets"]) == 1

    rv = client.get("/datasets?key1=[val21]")
    assert rv.status_code == 200
    assert len(json.loads(rv.data)["datasets"]) == 3

    rv = client.get("/datasets?key2=[val23,val24]&akey=x")
    assert rv.status_code == 200
    assert len(json.loads(rv.data)["datasets"]) == 2


def test_get_datasets_unknown_key(client, use_testdb):
    rv = client.get("/datasets?UNKNOWN=val21")
    assert rv.status_code == 400


def test_get_singleband_greyscale(client, use_testdb, raster_file_xyz):
    import terracotta

    settings = terracotta.get_settings()

    x, y, z = raster_file_xyz
    rv = client.get(f"/singleband/val11/x/val12/{z}/{x}/{y}.png")
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE


def test_get_singleband_extra_args(client, use_testdb, raster_file_xyz):
    import terracotta

    settings = terracotta.get_settings()

    x, y, z = raster_file_xyz
    rv = client.get(f"/singleband/val11/x/val12/{z}/{x}/{y}.png?foo=bar&baz=quz")
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE


def test_get_singleband_cmap(client, use_testdb, raster_file_xyz):
    import terracotta

    settings = terracotta.get_settings()

    x, y, z = raster_file_xyz
    rv = client.get(f"/singleband/val11/x/val12/{z}/{x}/{y}.png?colormap=jet")
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE


def test_get_singleband_preview(client, use_testdb):
    import terracotta

    settings = terracotta.get_settings()

    rv = client.get("/singleband/val11/x/val12/preview.png?colormap=jet")
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE


def urlsafe_json(payload):
    payload_json = json.dumps(payload)
    return urllib.parse.quote_plus(payload_json, safe=r',.[]{}:"')


def test_get_singleband_explicit_cmap(client, use_testdb, raster_file_xyz):
    import terracotta

    settings = terracotta.get_settings()

    x, y, z = raster_file_xyz
    explicit_cmap = {1: (0, 0, 0), 2.0: (255, 255, 255, 20), 3: "#ffffff", 4: "abcabc"}

    rv = client.get(
        f"/singleband/val11/x/val12/{z}/{x}/{y}.png?colormap=explicit"
        f"&explicit_color_map={urlsafe_json(explicit_cmap)}"
    )
    assert rv.status_code == 200, rv.data.decode("utf-8")

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE


def test_get_singleband_explicit_cmap_invalid(client, use_testdb, raster_file_xyz):
    x, y, z = raster_file_xyz
    explicit_cmap = {1: (0, 0, 0), 2: (255, 255, 255), 3: "#ffffff", 4: "abcabc"}

    rv = client.get(
        f"/singleband/val11/x/val12/{z}/{x}/{y}.png?"
        f"explicit_color_map={urlsafe_json(explicit_cmap)}"
    )
    assert rv.status_code == 400

    rv = client.get(
        f"/singleband/val11/x/val12/{z}/{x}/{y}.png?colormap=jet"
        f"&explicit_color_map={urlsafe_json(explicit_cmap)}"
    )
    assert rv.status_code == 400

    rv = client.get(f"/singleband/val11/x/val12/{z}/{x}/{y}.png?colormap=explicit")
    assert rv.status_code == 400

    explicit_cmap[3] = "omgomg"
    rv = client.get(
        f"/singleband/val11/x/val12/{z}/{x}/{y}.png?colormap=explicit"
        f"&explicit_color_map={urlsafe_json(explicit_cmap)}"
    )
    assert rv.status_code == 400

    explicit_cmap = [(255, 255, 255)]
    rv = client.get(
        f"/singleband/val11/x/val12/{z}/{x}/{y}.png?colormap=explicit"
        f"&explicit_color_map={urlsafe_json(explicit_cmap)}"
    )
    assert rv.status_code == 400

    rv = client.get(
        f"/singleband/val11/x/val12/{z}/{x}/{y}.png?colormap=explicit"
        f"&explicit_color_map=foo"
    )
    assert rv.status_code == 400


def test_get_singleband_stretch(client, use_testdb, raster_file_xyz):
    import terracotta

    settings = terracotta.get_settings()

    x, y, z = raster_file_xyz

    for stretch_range in (
        "[0,1]",
        "[0,null]",
        "[null, 1]",
        "[null,null]",
        "null",
        '["p2","p98"]',
        '[0, "p98"]',
    ):
        rv = client.get(
            f"/singleband/val11/x/val12/{z}/{x}/{y}.png?stretch_range={stretch_range}"
        )
        assert rv.status_code == 200

        img = Image.open(BytesIO(rv.data))
        assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE


def test_get_singleband_out_of_bounds(client, use_testdb):
    import terracotta

    settings = terracotta.get_settings()

    x, y, z = (0, 0, 10)
    rv = client.get(f"/singleband/val11/x/val12/{z}/{x}/{y}.png")
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE
    assert np.all(np.asarray(img) == 0)


def test_get_singleband_unknown_cmap(client, use_testdb, raster_file_xyz):
    x, y, z = raster_file_xyz
    rv = client.get(f"/singleband/val11/x/val12/{z}/{x}/{y}.png?colormap=UNKNOWN")
    assert rv.status_code == 400


def test_get_rgb(client, use_testdb, raster_file_xyz):
    import terracotta

    settings = terracotta.get_settings()

    x, y, z = raster_file_xyz
    rv = client.get(f"/rgb/val21/x/{z}/{x}/{y}.png?r=val22&g=val23&b=val24")
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == (*settings.DEFAULT_TILE_SIZE, 3)


def test_get_rgb_preview(client, use_testdb):
    import terracotta

    settings = terracotta.get_settings()

    rv = client.get("/rgb/val21/x/preview.png?r=val22&g=val23&b=val24")
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == (*settings.DEFAULT_TILE_SIZE, 3)


def test_get_rgb_extra_args(client, use_testdb, raster_file_xyz):
    import terracotta

    settings = terracotta.get_settings()

    x, y, z = raster_file_xyz
    rv = client.get(
        f"/rgb/val21/x/{z}/{x}/{y}.png?r=val22&g=val23&b=val24&foo=bar&baz=quz"
    )
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == (*settings.DEFAULT_TILE_SIZE, 3)


def test_get_rgb_stretch(client, use_testdb, raster_file_xyz):
    import terracotta

    settings = terracotta.get_settings()

    x, y, z = raster_file_xyz

    for stretch_range in (
        "[0,10000]",
        "[1.0e%2B01,1.0e%2B04]",
        '["p2","p98"]',
        "[0,null]",
        "[null, 10000]",
        "[null,null]",
        "null",
    ):
        rv = client.get(
            f"/rgb/val21/x/{z}/{x}/{y}.png?r=val22&g=val23&b=val24&"
            f"r_range={stretch_range}&b_range={stretch_range}&g_range={stretch_range}"
        )
        assert rv.status_code == 200, rv.data

        img = Image.open(BytesIO(rv.data))
        assert np.asarray(img).shape == (*settings.DEFAULT_TILE_SIZE, 3)


def test_get_compute(client, use_testdb, raster_file_xyz):
    import terracotta

    settings = terracotta.get_settings()

    # default tile size
    x, y, z = raster_file_xyz
    rv = client.get(
        f"/compute/val21/x/{z}/{x}/{y}.png"
        "?expression=v1*v2&v1=val22&v2=val23"
        "&stretch_range=[0,10000]"
    )
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE

    # custom tile size
    rv = client.get(
        f"/compute/val21/x/{z}/{x}/{y}.png"
        "?expression=v1*v2&v1=val22&v2=val23"
        "&stretch_range=[0,10000]"
        "&tile_size=[128,128]"
    )
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == (128, 128)


def test_get_compute_preview(client, use_testdb):
    import terracotta

    settings = terracotta.get_settings()

    rv = client.get(
        "/compute/val21/x/preview.png"
        "?expression=v1*v2&v1=val22&v2=val23"
        "&stretch_range=[0,10000]"
    )
    assert rv.status_code == 200

    img = Image.open(BytesIO(rv.data))
    assert np.asarray(img).shape == settings.DEFAULT_TILE_SIZE


def test_get_compute_invalid(client, use_testdb, raster_file_xyz):
    x, y, z = raster_file_xyz

    # too few keys
    rv = client.get(
        f"/compute/val21/{z}/{x}/{y}.png"
        "?expression=v1*v2&v1=val22&v2=val23"
        "&stretch_range=[0,10000]"
    )
    assert rv.status_code == 400

    # invalid expression
    rv = client.get(
        "/compute/val21/x/preview.png"
        '?expression=__builtins__["dir"](v1)&v1=val22'
        "&stretch_range=[0,10000]"
    )
    assert rv.status_code == 400

    # no stretch range
    rv = client.get(
        f"/compute/val21/x/{z}/{x}/{y}.png" "?expression=v1*v2&v1=val22&v2=val23"
    )
    assert rv.status_code == 400

    # no expression
    rv = client.get(f"/compute/val21/x/{z}/{x}/{y}.png" "?stretch_range=[0,10000)")
    assert rv.status_code == 400

    # missing operand
    rv = client.get(
        f"/compute/val21/x/{z}/{x}/{y}.png"
        "?expression=v1*v2"
        "&stretch_range=[0,10000)"
    )
    assert rv.status_code == 400

    # invalid stretch range (syntax)
    rv = client.get(
        f"/compute/val21/x/{z}/{x}/{y}.png"
        "?expression=v1*v2&v1=val22&v2=val23"
        "&stretch_range=[0,10000)"
    )
    assert rv.status_code == 400

    # invalid stretch range (value)
    rv = client.get(
        f"/compute/val21/x/{z}/{x}/{y}.png"
        "?expression=v1*v2&v1=val22&v2=val23"
        "&stretch_range=[10000,0]"
    )
    assert rv.status_code == 400


def test_get_colormap(client):
    rv = client.get("/colormap?stretch_range=[0,1]&num_values=100")
    assert rv.status_code == 200
    assert len(json.loads(rv.data)["colormap"]) == 100


def test_get_colormap_invalid(client):
    rv = client.get("/colormap?stretch_range=[0,1")
    assert rv.status_code == 400


def test_get_colormap_extra_args(client):
    rv = client.get("/colormap?stretch_range=[0,1]&num_values=100&foo=bar&baz=quz")
    assert rv.status_code == 200
    assert len(json.loads(rv.data)["colormap"]) == 100


def test_get_spec(client):
    from terracotta import __version__

    rv = client.get("/swagger.json")
    assert rv.status_code == 200
    assert json.loads(rv.data)
    assert __version__ in rv.data.decode("utf-8")

    rv = client.get("/apidoc")
    assert rv.status_code == 200
    assert b"Terracotta" in rv.data


@pytest.mark.parametrize(
    "stretch_range_params",
    [
        ['["s2", "p98"]', "Percentile format is"],
        ['["pp2", "p98"]', "Percentile format is"],
        ['["p", "p98"]', "Percentile format is"],
        ['["2", "p98"]', "Percentile format is"],
        ['["[]", "p98"]', "Percentile format is"],
    ],
)
def test_get_rgb_invalid_stretch_percentile(
    debug_client, use_testdb, raster_file_xyz, stretch_range_params
):

    x, y, z = raster_file_xyz
    stretch_range = stretch_range_params[0]
    with pytest.raises(marshmallow.ValidationError) as exc:
        debug_client.get(
            f"/rgb/val21/x/{z}/{x}/{y}.png?r=val22&g=val23&b=val24&"
            f"r_range={stretch_range}&b_range={stretch_range}&g_range={stretch_range}"
        )
    assert stretch_range_params[1] in str(exc.value)



import zlib

import numpy as np


def test_get_size():
    from terracotta.cache import CompressedLFUCache

    tile_shape = (256, 256)
    data = zlib.compress(np.ones(tile_shape), 9)
    mask = zlib.compress(np.zeros(tile_shape), 9)
    size = CompressedLFUCache._get_size((data, mask, "float64", tile_shape))
    assert 1450 < size < 1550



"""handlers/metadata.py

Handle /metadata API endpoint.
"""

from typing import Mapping, Sequence, Dict, Any, Union, List, Optional
from collections import OrderedDict

from terracotta import get_settings, get_driver
from terracotta.exceptions import InvalidArgumentsError
from terracotta.profile import trace


def filter_metadata(
    metadata: Dict[str, Any], columns: Optional[List[str]]
) -> Dict[str, Any]:
    """Filter metadata by columns, if given"""
    assert (
        columns is None or len(columns) > 0
    ), "columns must either be a non-empty list or None"

    if columns:
        metadata = {c: metadata[c] for c in columns}

    return metadata


@trace("metadata_handler")
def metadata(
    columns: Optional[List[str]], keys: Union[Sequence[str], Mapping[str, str]]
) -> Dict[str, Any]:
    """Returns all metadata for a single dataset"""
    settings = get_settings()
    driver = get_driver(settings.DRIVER_PATH, provider=settings.DRIVER_PROVIDER)
    metadata = filter_metadata(driver.get_metadata(keys), columns)
    metadata["keys"] = OrderedDict(zip(driver.key_names, keys))
    return metadata


@trace("multiple_metadata_handler")
def multiple_metadata(
    columns: Optional[List[str]], datasets: List[List[str]]
) -> List[Dict[str, Any]]:
    """Returns all metadata for multiple datasets"""
    settings = get_settings()
    driver = get_driver(settings.DRIVER_PATH, provider=settings.DRIVER_PROVIDER)
    key_names = driver.key_names

    if len(datasets) > settings.MAX_POST_METADATA_KEYS:
        raise InvalidArgumentsError(
            f"Maximum number of keys exceeded ({settings.MAX_POST_METADATA_KEYS}). "
            f"This limit can be configured in the server settings."
        )

    out = []
    with driver.connect():
        for dataset in datasets:
            metadata = filter_metadata(driver.get_metadata(dataset), columns)
            metadata["keys"] = OrderedDict(zip(key_names, dataset))
            out.append(metadata)

    return out



"""drivers/terracotta_driver.py

The driver to interact with.
"""

import contextlib
from collections import OrderedDict
from typing import (
    Any,
    Collection,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import terracotta
from terracotta import exceptions
from terracotta.drivers.base_classes import (
    KeysType,
    MetaStore,
    MultiValueKeysType,
    RasterStore,
)

ExtendedKeysType = Union[Sequence[str], Mapping[str, str]]
ExtendedMultiValueKeysType = Union[Sequence[str], Mapping[str, Union[str, List[str]]]]
T = TypeVar("T")


def squeeze(iterable: Collection[T]) -> T:
    assert len(iterable) == 1
    return next(iter(iterable))


class TerracottaDriver:
    """Terracotta driver object used to retrieve raster tiles and metadata.

    Do not instantiate directly, use :func:`terracotta.get_driver` instead.
    """

    def __init__(self, meta_store: MetaStore, raster_store: RasterStore) -> None:
        self.meta_store = meta_store
        self.raster_store = raster_store

        settings = terracotta.get_settings()
        self.LAZY_LOADING_MAX_SHAPE: Tuple[int, int] = settings.LAZY_LOADING_MAX_SHAPE

    @property
    def db_version(self) -> str:
        """Terracotta version used to create the meta store.

        Returns:

            A str specifying the version of Terracotta that was used to create the meta store.

        """
        return self.meta_store.db_version

    @property
    def key_names(self) -> Tuple[str, ...]:
        """Get names of all keys defined by the meta store.

        Returns:

            A tuple defining the key names and order.

        """
        return self.meta_store.key_names

    def create(
        self,
        keys: Sequence[str],
        *,
        key_descriptions: Optional[Mapping[str, str]] = None,
    ) -> None:
        """Create a new, empty metadata store.

        Arguments:

            keys: A sequence defining the key names and order.
            key_descriptions: A mapping from key name to a human-readable
                description of what the key encodes.

        """
        self.meta_store.create(keys=keys, key_descriptions=key_descriptions)

    def connect(self, verify: bool = True) -> contextlib.AbstractContextManager:
        """Context manager to connect to the metastore and clean up on exit.

        This allows you to pool interactions with the metastore to prevent possibly
        expensive reconnects, or to roll back several interactions if one of them fails.

        Arguments:

            verify: Whether to verify the metastore (primarily its version) when connecting.
                Should be `true` unless absolutely necessary, such as when instantiating the
                metastore during creation of it.

        Note:

            Make sure to call :meth:`create` on a fresh metastore before using this method.

        Example:

            >>> import terracotta as tc
            >>> driver = tc.get_driver('tc.sqlite')
            >>> with driver.connect():
            ...     for keys, dataset in datasets.items():
            ...         # connection will be kept open between insert operations
            ...         driver.insert(keys, dataset)

        """
        return self.meta_store.connect(verify=verify)

    def get_keys(self) -> OrderedDict:
        """Get all known keys and their fulltext descriptions.

        Returns:

            An :class:`~collections.OrderedDict` in the form
            ``{key_name: key_description}``

        """
        return self.meta_store.get_keys()

    def get_datasets(
        self,
        where: Optional[ExtendedMultiValueKeysType] = None,
        page: int = 0,
        limit: Optional[int] = None,
    ) -> Dict[Tuple[str, ...], Any]:
        """Get all known dataset key combinations matching the given constraints,
        and a path to retrieve the data (dependent on the raster store).

        Arguments:

            where: A mapping from key name to key value constraint(s)
            page: A pagination parameter, skips first page * limit results
            limit: A pagination parameter, max number of results to return

        Returns:

            A :class:`dict` mapping from key sequence tuple to dataset path.

        """
        return self.meta_store.get_datasets(
            where=self._standardize_multi_value_keys(where, requires_all_keys=False),
            page=page,
            limit=limit,
        )

    def get_metadata(self, keys: ExtendedKeysType) -> Dict[str, Any]:
        """Return all stored metadata for given keys.

        Arguments:

            keys: Keys of the requested dataset. Can either be given as a sequence of key values,
                or as a mapping ``{key_name: key_value}``.

        Returns:

            A :class:`dict` with the values

            - ``range``: global minimum and maximum value in dataset
            - ``bounds``: physical bounds covered by dataset in latitude-longitude projection
            - ``convex_hull``: GeoJSON shape specifying total data coverage in latitude-longitude
              projection
            - ``percentiles``: array of pre-computed percentiles from 1% through 99%
            - ``mean``: global mean
            - ``stdev``: global standard deviation
            - ``metadata``: any additional client-relevant metadata

        """
        keys = self._standardize_keys(keys)

        with self.meta_store.connect():
            metadata = self.meta_store.get_metadata(keys)

            if metadata is None:
                # metadata is not computed yet, trigger lazy loading
                dataset = self.get_datasets(keys)
                if not dataset:
                    raise exceptions.DatasetNotFoundError("No dataset found")

                path = squeeze(dataset.values())
                metadata = self.compute_metadata(
                    path, max_shape=self.LAZY_LOADING_MAX_SHAPE
                )

                try:
                    self.insert(keys, path, metadata=metadata)
                except exceptions.DatabaseNotWritableError as exc:
                    raise exceptions.DatabaseNotWritableError(
                        "Lazy loading requires a writable database"
                    ) from exc

                # ensure standardized/consistent output (types and floating point precision)
                metadata = self.meta_store.get_metadata(keys)
                assert metadata is not None

        return metadata

    def insert(
        self,
        keys: ExtendedKeysType,
        path: str,
        *,
        override_path: Optional[str] = None,
        metadata: Optional[Mapping[str, Any]] = None,
        skip_metadata: bool = False,
    ) -> None:
        """Register a new dataset. Used to populate meta store.

        Arguments:

            keys: Keys of the dataset. Can either be given as a sequence of key values, or
                as a mapping ``{key_name: key_value}``.
            path: Path to access dataset (driver dependent).
            override_path: If given, this path will be inserted into the meta store
                instead of the one used to load the dataset.
            metadata: Metadata dict for the dataset. If not given, metadata will be computed
                via :meth:`compute_metadata`.
            skip_metadata: If True, will skip metadata computation (will be computed
                during first request instead). Has no effect if ``metadata`` argument is given.

        """
        keys = self._standardize_keys(keys)

        if metadata is None and not skip_metadata:
            metadata = self.compute_metadata(path)

        self.meta_store.insert(keys=keys, path=override_path or path, metadata=metadata)

    def delete(self, keys: ExtendedKeysType) -> None:
        """Remove a dataset from the meta store.

        Arguments:

            keys:  Keys of the dataset. Can either be given as a sequence of key values, or
                as a mapping ``{key_name: key_value}``.

        """
        keys = self._standardize_keys(keys)

        self.meta_store.delete(keys)

    def get_raster_tile(
        self,
        keys: ExtendedKeysType,
        *,
        tile_bounds: Optional[Sequence[float]] = None,
        tile_size: Sequence[int] = (256, 256),
        preserve_values: bool = False,
        asynchronous: bool = False,
    ) -> Any:
        """Load a raster tile with given keys and bounds.

        Arguments:

            keys: Key sequence identifying the dataset to load tile from.
            tile_bounds: Physical bounds of the tile to read, in Web Mercator projection (EPSG3857).
                Reads the whole dataset if not given.
            tile_size: Shape of the output array to return. Must be two-dimensional.
                Defaults to :attr:`~terracotta.config.TerracottaSettings.DEFAULT_TILE_SIZE`.
            preserve_values: Whether to preserve exact numerical values (e.g. when reading
                categorical data). Sets all interpolation to nearest neighbor.
            asynchronous: If given, the tile will be read asynchronously in a separate thread.
                This function will return immediately with a :class:`~concurrent.futures.Future`
                that can be used to retrieve the result.

        Returns:

            Requested tile as :class:`~numpy.ma.MaskedArray` of shape ``tile_size`` if
            ``asynchronous=False``, otherwise a :class:`~concurrent.futures.Future` containing
            the result.

        """
        path = squeeze(self.get_datasets(keys).values())

        return self.raster_store.get_raster_tile(
            path=path,
            tile_bounds=tile_bounds,
            tile_size=tile_size,
            preserve_values=preserve_values,
            asynchronous=asynchronous,
        )

    def compute_metadata(
        self,
        path: str,
        *,
        extra_metadata: Optional[Any] = None,
        use_chunks: Optional[bool] = None,
        max_shape: Optional[Sequence[int]] = None,
    ) -> Dict[str, Any]:
        """Compute metadata for a dataset.

        Arguments:

            path: Path identifing the dataset.
            extra_metadata: Any additional metadata that will be returned as is
                in the result, under the `metadata` key.
            use_chunks: Whether to load the dataset in chunks, when computing.
                Useful if the dataset is too large to fit in memory.
                Mutually exclusive with `max_shape`.
            max_shape: If dataset is larger than this shape, it will be downsampled
                while loading. Useful if the dataset is too large to fit in memory.
                Mutually exclusive with `use_chunks`.

        Returns:

            A :class:`dict` with the values

            - ``range``: global minimum and maximum value in dataset
            - ``bounds``: physical bounds covered by dataset in latitude-longitude projection
            - ``convex_hull``: GeoJSON shape specifying total data coverage in latitude-longitude
              projection
            - ``percentiles``: array of pre-computed percentiles from 1% through 99%
            - ``mean``: global mean
            - ``stdev``: global standard deviation
            - ``metadata``: any additional client-relevant metadata

        """
        return self.raster_store.compute_metadata(
            path=path,
            extra_metadata=extra_metadata,
            use_chunks=use_chunks,
            max_shape=max_shape,
        )

    def _standardize_keys(
        self, keys: ExtendedKeysType, requires_all_keys: bool = True
    ) -> KeysType:
        return self._ensure_keys_as_dict(keys, requires_all_keys)

    def _standardize_multi_value_keys(
        self, keys: Optional[ExtendedMultiValueKeysType], requires_all_keys: bool = True
    ) -> MultiValueKeysType:
        return self._ensure_keys_as_dict(keys, requires_all_keys)

    def _ensure_keys_as_dict(
        self,
        keys: Union[ExtendedKeysType, Optional[MultiValueKeysType]],
        requires_all_keys: bool = True,
    ) -> Dict[str, Any]:
        if requires_all_keys and (keys is None or len(keys) != len(self.key_names)):
            raise exceptions.InvalidKeyError(
                f"Got wrong number of keys (available keys: {self.key_names})"
            )

        if isinstance(keys, Mapping):
            keys = dict(keys.items())
        elif isinstance(keys, Sequence):
            keys = dict(zip(self.key_names, keys))
        elif keys is None:
            keys = {}
        else:
            raise exceptions.InvalidKeyError(
                "Encountered unknown key type, expected Mapping or Sequence"
            )

        unknown_keys = set(keys) - set(self.key_names)
        if unknown_keys:
            raise exceptions.InvalidKeyError(
                f"Encountered unrecognized keys {unknown_keys} (available keys: {self.key_names})"
            )

        return keys

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(\n"
            f"    meta_store={self.meta_store!r},\n"
            f"    raster_store={self.raster_store!r}\n"
            ")"
        )



import pytest
from click.testing import CliRunner

import os
import shutil

TEST_CASES = (
    {  # basic
        "filenames": ["foo.tif"],
        "input_pattern": "{name}.tif",
        "expected_keys": ["name"],
        "expected_datasets": [("foo",)],
    },
    {  # two keys
        "filenames": ["S2_20180101_B04.tif"],
        "input_pattern": "{sensor}_{date}_B04.tif",
        "expected_keys": ["sensor", "date"],
        "expected_datasets": [("S2", "20180101")],
    },
    {  # subfolder
        "filenames": ["S2/20180101_B04.tif"],
        "input_pattern": "{sensor}/{date}_B04.tif",
        "expected_keys": ["sensor", "date"],
        "expected_datasets": [("S2", "20180101")],
    },
    {  # wildcard magic
        "filenames": ["S2/20180101_B04.tif", "S2/20180101_B05_median.tif"],
        "input_pattern": "{sensor}/{date}_{band}{}.tif",
        "expected_keys": ["sensor", "date", "band"],
        "expected_datasets": [("S2", "20180101", "B04"), ("S2", "20180101", "B05")],
    },
    {  # keys occuring more than once
        "filenames": ["S2/20180101/S2_20180101_B04.tif"],
        "input_pattern": "{sensor}/{date}/{sensor}_{date}_{band}.tif",
        "expected_keys": ["sensor", "date", "band"],
        "expected_datasets": [("S2", "20180101", "B04")],
    },
    {  # {} in filename
        "filenames": ["bar.tif", "{foo}.tif"],
        "input_pattern": "{{{name}}}.tif",
        "expected_keys": ["name"],
        "expected_datasets": [("foo",)],
    },
    {  # case-sensitivity
        "filenames": ["bAr.tif", "FOO.tif"],
        "input_pattern": "{nAmE}.tif",
        "expected_keys": ["nAmE"],
        "expected_datasets": [("bAr",), ("FOO",)],
    },
    {  # unicode path with space
        "filenames": ["süßer ordner/foo.tif"],
        "input_pattern": "{}/{name}.tif",
        "expected_keys": ["name"],
        "expected_datasets": [("foo",)],
    },
    {  # unicode key
        "filenames": ["günther.tif"],
        "input_pattern": "{bärbel}.tif",
        "expected_keys": ["bärbel"],
        "expected_datasets": [("günther",)],
    },
)

INVALID_TEST_CASES = (
    {  # no matching files
        "filenames": [],
        "input_pattern": "{key}.tif",
        "error_contains": "matches no files",
    },
    {  # duplicate keys in different folders
        "filenames": ["dir1/foo.tif", "dir2/foo.tif"],
        "input_pattern": "{}/{name}.tif",
        "error_contains": "duplicate keys",
    },
    {  # duplicate keys through wildcard
        "filenames": ["S2_B04.tif", "S2_20180101_B04.tif"],
        "input_pattern": "{sensor}_{}.tif",
        "error_contains": "duplicate keys",
    },
    {  # no groups in pattern
        "filenames": [],
        "input_pattern": "notafile.tif",
        "error_contains": "at least one placeholder",
    },
    {  # only wildcards in pattern
        "filenames": [],
        "input_pattern": "{}.tif",
        "error_contains": "at least one placeholder",
    },
    {  # stray {
        "filenames": [],
        "input_pattern": "notafile{.tif",
        "error_contains": "invalid pattern",
    },
    {  # invalid placeholder name
        "filenames": ["foo.tif"],
        "input_pattern": "{(foo)}.tif",
        "error_contains": "must be alphanumeric",
    },
)


@pytest.fixture()
def tmpworkdir(tmpdir):
    orig_dir = os.getcwd()
    try:
        os.chdir(tmpdir)
        yield tmpdir
    finally:
        os.chdir(orig_dir)


def test_ingest(raster_file, tmpdir):
    from terracotta.scripts import cli

    outfile = tmpdir / "out.sqlite"
    input_pattern = str(raster_file.dirpath("{name}.tif"))

    runner = CliRunner()
    result = runner.invoke(cli.cli, ["ingest", input_pattern, "-o", str(outfile)])
    assert result.exit_code == 0
    assert outfile.check()

    from terracotta import get_driver

    driver = get_driver(str(outfile), provider="sqlite")
    assert driver.key_names == ("name",)
    assert driver.get_datasets() == {("img",): str(raster_file)}


def test_ingest_append(raster_file, tmpworkdir):
    from terracotta.scripts import cli

    for infile in ("dir1/img1.tif", "dir2/img2.tif"):
        temp_infile = tmpworkdir / infile
        os.makedirs(temp_infile.dirpath(), exist_ok=True)
        shutil.copy(raster_file, temp_infile)

    outfile = tmpworkdir / "out.sqlite"

    runner = CliRunner()
    result = runner.invoke(cli.cli, ["ingest", "dir1/{name}.tif", "-o", str(outfile)])
    assert result.exit_code == 0
    assert outfile.check()

    result = runner.invoke(cli.cli, ["ingest", "dir2/{name}.tif", "-o", str(outfile)])
    assert result.exit_code == 0
    assert outfile.check()

    from terracotta import get_driver

    driver = get_driver(str(outfile), provider="sqlite")
    assert driver.key_names == ("name",)
    assert all((ds,) in driver.get_datasets() for ds in ("img1", "img2"))


@pytest.mark.parametrize("skip_existing", [True, False])
def test_reingest(skip_existing, raster_file, tmpworkdir):
    from terracotta.scripts import cli
    from terracotta import get_driver

    same_name = "myimage"
    infiles = [
        tmpworkdir / p
        for p in [
            f"dir1/{same_name}.tif",
            f"dir2/{same_name}.tif",
            f"dir3/{same_name}.tif",
        ]
    ]
    for temp_infile in infiles:
        os.makedirs(temp_infile.dirpath(), exist_ok=True)
        shutil.copy(raster_file, temp_infile)

    outfile = tmpworkdir / "out.sqlite"

    def _assert_datasets_equal(datasets):
        assert outfile.check()
        driver = get_driver(str(outfile), provider="sqlite")
        existing_datasets = driver.get_datasets()
        assert len(existing_datasets) == 1
        assert existing_datasets == datasets

    runner = CliRunner()
    args = ["ingest", "dir1/{name}.tif", "-o", str(outfile)]
    result = runner.invoke(cli.cli, args)
    assert result.exit_code == 0
    _assert_datasets_equal({(same_name,): str(infiles[0])})

    args = ["ingest", "dir2/{name}.tif", "-o", str(outfile)]
    if skip_existing:
        args.append("--skip-existing")
    result = runner.invoke(cli.cli, args)
    assert result.exit_code == 0
    if skip_existing:
        _assert_datasets_equal({(same_name,): str(infiles[0])})
    else:
        _assert_datasets_equal({(same_name,): str(infiles[1])})

    args = ["ingest", "dir3/{name}.tif", "-o", str(outfile)]
    if skip_existing:
        args.append("--skip-existing")
    result = runner.invoke(cli.cli, args)
    assert result.exit_code == 0
    if skip_existing:
        _assert_datasets_equal({(same_name,): str(infiles[0])})
    else:
        _assert_datasets_equal({(same_name,): str(infiles[2])})


def test_ingest_append_invalid(raster_file, tmpworkdir):
    from terracotta.scripts import cli

    for infile in ("dir1/img1.tif", "dir2/img2.tif"):
        temp_infile = tmpworkdir / infile
        os.makedirs(temp_infile.dirpath(), exist_ok=True)
        shutil.copy(raster_file, temp_infile)

    outfile = tmpworkdir / "out.sqlite"

    runner = CliRunner()
    result = runner.invoke(cli.cli, ["ingest", "dir1/{name}.tif", "-o", str(outfile)])
    assert result.exit_code == 0
    assert outfile.check()

    result = runner.invoke(cli.cli, ["ingest", "{dir}/{name}.tif", "-o", str(outfile)])
    assert result.exit_code != 0
    assert "incompatible key names" in result.output


@pytest.mark.parametrize("case", TEST_CASES)
@pytest.mark.parametrize("abspath", [True, False])
def test_ingest_pattern(case, abspath, raster_file, tmpworkdir):
    from terracotta.scripts import cli

    for infile in case["filenames"]:
        temp_infile = tmpworkdir / infile
        os.makedirs(temp_infile.dirpath(), exist_ok=True)
        shutil.copy(raster_file, temp_infile)

    outfile = tmpworkdir / "out.sqlite"

    if abspath:
        input_pattern = os.path.abspath(tmpworkdir / case["input_pattern"])
    else:
        input_pattern = case["input_pattern"]

    runner = CliRunner()
    result = runner.invoke(cli.cli, ["ingest", input_pattern, "-o", str(outfile)])
    assert result.exit_code == 0, result.output
    assert outfile.check()

    from terracotta import get_driver

    driver = get_driver(str(outfile), provider="sqlite")
    assert driver.key_names == tuple(case["expected_keys"])
    assert all(ds in driver.get_datasets() for ds in case["expected_datasets"])


@pytest.mark.parametrize("case", INVALID_TEST_CASES)
def test_ingest_invalid_pattern(case, raster_file, tmpworkdir):
    from terracotta.scripts import cli

    for infile in case["filenames"]:
        temp_infile = tmpworkdir / infile
        os.makedirs(temp_infile.dirpath(), exist_ok=True)
        shutil.copy(raster_file, temp_infile)

    outfile = tmpworkdir / "out.sqlite"
    input_pattern = case["input_pattern"]

    runner = CliRunner()
    result = runner.invoke(cli.cli, ["ingest", input_pattern, "-o", str(outfile)])
    assert result.exit_code != 0
    assert case["error_contains"].lower() in result.output.lower()


def test_ingest_rgb_key(raster_file, tmpdir):
    from terracotta.scripts import cli

    outfile = tmpdir / "out.sqlite"
    input_pattern = str(raster_file.dirpath("{rgb}m{foo}.tif"))

    runner = CliRunner()
    result = runner.invoke(
        cli.cli, ["ingest", input_pattern, "-o", str(outfile), "--rgb-key", "rgb"]
    )
    assert result.exit_code == 0
    assert outfile.check()

    from terracotta import get_driver

    driver = get_driver(str(outfile), provider="sqlite")
    assert driver.key_names == ("foo", "rgb")
    assert driver.get_datasets() == {("g", "i"): str(raster_file)}


def test_ingest_invalid_rgb_key(raster_file, tmpdir):
    from terracotta.scripts import cli

    outfile = tmpdir / "out.sqlite"
    input_pattern = str(raster_file.dirpath("{rgb}m{foo}.tif"))

    runner = CliRunner()
    result = runner.invoke(
        cli.cli, ["ingest", input_pattern, "-o", str(outfile), "--rgb-key", "bar"]
    )
    assert result.exit_code != 0
    assert not outfile.check()



import os
import warnings
import traceback

import rasterio
import numpy as np

from click.testing import CliRunner

import pytest


def format_exception(result):
    return "".join(traceback.format_exception(*result.exc_info))


@pytest.fixture()
def tiny_raster_file(unoptimized_raster_file, tmpdir_factory):
    tmpdir = tmpdir_factory.mktemp("tiny-raster")
    outfile = tmpdir / "tiny.tif"
    with rasterio.open(str(unoptimized_raster_file)) as src:
        profile = src.profile.copy()
        profile.update(width=100, height=100, blockxsize=256, blockysize=256)

        with rasterio.open(str(outfile), "w", **profile) as dst:
            dst.write(src.read()[:100, :100])

    yield outfile


@pytest.mark.parametrize("in_memory", [True, None, False])
@pytest.mark.parametrize("reproject", [True, False])
@pytest.mark.parametrize("compression", ["auto", "lzw", "none"])
@pytest.mark.parametrize("nproc", [None, 1, 2, -1])
def test_optimize_rasters(
    unoptimized_raster_file, tmpdir, in_memory, reproject, compression, nproc
):
    from terracotta.cog import validate
    from terracotta.scripts import cli

    input_pattern = str(unoptimized_raster_file.dirpath("*.tif"))
    outfile = tmpdir / unoptimized_raster_file.basename

    runner = CliRunner()

    flags = ["--compression", compression, "-q"]

    if in_memory is not None:
        flags.append("--in-memory" if in_memory else "--no-in-memory")

    if reproject:
        flags.append("--reproject")

    if nproc is not None:
        flags.append(f"--nproc={nproc}")

    result = runner.invoke(
        cli.cli, ["optimize-rasters", input_pattern, "-o", str(tmpdir), *flags]
    )

    assert result.exit_code == 0, format_exception(result)
    assert outfile.check()

    # validate files
    assert not validate(str(unoptimized_raster_file))
    assert validate(str(outfile))

    if reproject:
        return

    # check for data integrity
    with rasterio.open(str(unoptimized_raster_file)) as src1, rasterio.open(
        str(outfile)
    ) as src2:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "invalid value encountered.*")
            np.testing.assert_array_equal(src1.read(), src2.read())


def test_optimize_rasters_small(tiny_raster_file, tmpdir):
    from terracotta.cog import validate
    from terracotta.scripts import cli

    input_pattern = str(tiny_raster_file)
    outfile = tmpdir / tiny_raster_file.basename

    runner = CliRunner()
    result = runner.invoke(
        cli.cli, ["optimize-rasters", input_pattern, "-o", str(tmpdir)]
    )

    assert result.exit_code == 0, format_exception(result)
    assert outfile.check()

    # validate files
    # (small rasters don't need overviews, so input file is valid, too)
    assert validate(str(tiny_raster_file))
    assert validate(str(outfile))

    # check for data integrity
    with rasterio.open(str(tiny_raster_file)) as src1, rasterio.open(
        str(outfile)
    ) as src2:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "invalid value encountered.*")
            np.testing.assert_array_equal(src1.read(), src2.read())


def test_optimize_rasters_nofiles(tmpdir):
    from terracotta.scripts import cli

    input_pattern = str(tmpdir.dirpath("*.tif"))
    runner = CliRunner()
    result = runner.invoke(
        cli.cli, ["optimize-rasters", input_pattern, "-o", str(tmpdir), "-q"]
    )

    assert result.exit_code == 0
    assert "No files given" in result.output


def test_optimize_rasters_invalid(tmpdir):
    from terracotta.scripts import cli

    runner = CliRunner()
    result = runner.invoke(
        cli.cli, ["optimize-rasters", str(tmpdir), "-o", str(tmpdir), "-q"]
    )

    assert result.exit_code != 0
    assert "not a file" in result.output

    result = runner.invoke(
        cli.cli,
        [
            "optimize-rasters",
            str(tmpdir),
            "-o",
            str(tmpdir),
            "--overwrite",
            "--skip-existing",
        ],
    )
    assert result.exit_code != 0
    assert "mutually exclusive" in result.output


def test_optimize_rasters_multiband(tmpdir, unoptimized_raster_file):
    from terracotta.scripts import cli
    import rasterio

    with rasterio.open(str(unoptimized_raster_file)) as src:
        profile = src.profile.copy()
        data = src.read(1)

    profile["count"] = 3

    multiband_file = tmpdir.join(unoptimized_raster_file.basename)
    with rasterio.open(str(multiband_file), "w", **profile) as dst:
        dst.write(data, 1)
        dst.write(data, 2)
        dst.write(data, 3)

    input_pattern = str(multiband_file.dirpath("*.tif"))
    outfile = tmpdir / "co" / unoptimized_raster_file.basename

    runner = CliRunner()
    result = runner.invoke(
        cli.cli, ["optimize-rasters", input_pattern, "-o", str(tmpdir / "co")]
    )

    assert result.exit_code == 0
    assert "has more than one band" in result.output

    with rasterio.open(str(unoptimized_raster_file)) as src1, rasterio.open(
        str(outfile)
    ) as src2:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "invalid value encountered.*")
            np.testing.assert_array_equal(src1.read(), src2.read())


@pytest.mark.parametrize("extra_flag", ["skip-existing", "overwrite", None])
def test_reoptimize(tmpdir, unoptimized_raster_file, extra_flag):
    from terracotta.scripts import cli

    infile = str(unoptimized_raster_file.dirpath("*.tif"))
    outfile = tmpdir / "out.tif"

    # first time
    runner = CliRunner()
    args = ["optimize-rasters", infile, "-o", str(outfile)]
    result = runner.invoke(cli.cli, args)
    assert result.exit_code == 0
    ctime = os.path.getmtime(outfile)

    # second time
    args = ["optimize-rasters", infile, "-o", str(outfile)]
    if extra_flag:
        args.append(f"--{extra_flag}")

    result = runner.invoke(cli.cli, args)

    if extra_flag == "skip-existing":
        assert result.exit_code == 0
        assert os.path.getmtime(outfile) == ctime
    elif extra_flag == "overwrite":
        assert result.exit_code == 0
        assert os.path.getmtime(outfile) != ctime
    else:
        assert result.exit_code == 2


def _throw(*args):
    raise RuntimeError("A mock error is raised")


def test_exception_in_subprocess(unoptimized_raster_file, tmpdir, monkeypatch):
    from terracotta.scripts import cli

    monkeypatch.setattr(
        "terracotta.scripts.optimize_rasters._optimize_single_raster", _throw
    )

    args = [
        "optimize-rasters",
        str(unoptimized_raster_file),
        "-o",
        str(tmpdir / "foo.tif"),
        "--nproc",
        2,
    ]

    runner = CliRunner()
    result = runner.invoke(cli.cli, args)

    assert result.exit_code != 0
    assert "Error while optimizing file" in str(result.exception)



"""scripts/migrate.py

Migrate databases between Terracotta versions.
"""

from typing import Tuple

import click
import sqlalchemy as sqla

from terracotta import get_driver, __version__
from terracotta.migrations import MIGRATIONS
from terracotta.drivers.relational_meta_store import RelationalMetaStore


def parse_version(verstr: str) -> Tuple[int, ...]:
    """Convert 'v<major>.<minor>.<patch>' to (major, minor)"""
    components = verstr.split(".")
    components[0] = components[0].lstrip("v")
    return tuple(int(c) for c in components[:2])


def join_version(vertuple: Tuple[int, ...]) -> str:
    return "v" + ".".join(map(str, vertuple))


@click.argument("DATABASE", required=True)
@click.option("--from", "from_version", required=False, default=None)
@click.option("--to", "to_version", required=False, default=__version__)
@click.option("-y", "--yes", is_flag=True, help="Do not ask for confirmation.")
@click.command("migrate")
def migrate(database: str, to_version: str, from_version: str, yes: bool) -> None:
    """Migrate databases between Terracotta versions."""
    driver = get_driver(database)
    meta_store = driver.meta_store
    assert isinstance(meta_store, RelationalMetaStore)

    to_version_tuple = parse_version(to_version)
    tc_version_tuple = parse_version(__version__)

    if to_version_tuple > tc_version_tuple:
        raise ValueError(
            f"Unknown target version {join_version(to_version_tuple)} (this is {join_version(tc_version_tuple)}). Try upgrading terracotta."
        )

    if from_version is None:
        try:  # type: ignore
            with meta_store.connect(verify=False):
                from_version_tuple = parse_version(driver.db_version)
        except Exception as e:
            raise RuntimeError("Cannot determine database version.") from e
    else:
        from_version_tuple = parse_version(from_version)

    if from_version_tuple == to_version_tuple:
        click.echo("Already at target version, nothing to do.")
        return

    migration_chain = []
    current_version = from_version_tuple

    while current_version != to_version_tuple:
        if current_version not in MIGRATIONS:
            raise RuntimeError("Unexpected error")

        migration = MIGRATIONS[current_version]
        migration_chain.append(migration)
        current_version = migration.up_version

    click.echo("Upgrade path found\n")

    for migration in migration_chain:
        click.echo(
            f"{join_version(migration.down_version)} -> {join_version(migration.up_version)}"
        )

        for cmd in migration.upgrade_sql:
            click.echo(f"    {cmd}")

        click.echo("")

    click.echo(
        f"This will upgrade the database from {join_version(from_version_tuple)} -> {join_version(to_version_tuple)} and execute the above SQL commands."
    )

    if not yes:
        click.confirm("Continue?", abort=True)

    with meta_store.connect(verify=False) as conn:
        for migration in migration_chain:
            for cmd in migration.upgrade_sql:
                conn.execute(sqla.text(cmd))
